import io
import os
import unittest
from textwrap import dedent
from zipfile import ZipFile, BadZipFile

from bson import ObjectId

from retk import core, const
from retk.core.files import saver
from retk.core.files.importing.async_tasks.obsidian import ops
from retk.models.client import client
from . import utils


class ObsidianTest(unittest.IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        utils.set_env(".env.test.local")
        cls.orig_folder_data = {
            os.path.join("a", "test.md"): b"hello world",
            os.path.join("a", "a.md"): b"a",
            os.path.join("a", "v", "q文档.md"): "阿斯顿请问".encode("utf-8"),
        }
        cls.orig_data = {
            "test.md": b"hello world",
            "a.md": b"a",
            "q文档.md": "阿斯顿请问".encode("utf-8"),
        }

    async def asyncSetUp(self) -> None:
        await client.init()
        u, _ = await core.user.get_by_email(email=const.DEFAULT_USER["email"])
        self.uid = u["id"]

    async def asyncTearDown(self) -> None:
        await client.drop()

    @classmethod
    def tearDownClass(cls) -> None:
        utils.drop_env(".env.test.local")

    def test_unzip_folder(self):
        zip_bytes = ZipFile("test.zip", "w")
        for k, v in self.orig_folder_data.items():
            zip_bytes.writestr(k, v)
        zip_bytes.close()

        # unzip
        with open("test.zip", "rb") as f:
            extracted_files = ops.unzip_obsidian(f.read())
        for full_path, meta in extracted_files.md_full.items():
            self.assertEqual(full_path, meta.filepath)
            self.assertEqual(
                self.orig_folder_data[os.path.join("a", full_path)],
                meta.file,
                msg=str(self.orig_folder_data))
        for filename, meta in extracted_files.md.items():
            self.assertEqual(
                self.orig_data[filename],
                meta.file,
                msg=filename)
        os.remove("test.zip")

    def test_unzip_files(self):
        zip_bytes = ZipFile("test.zip", "w")
        for k, v in self.orig_data.items():
            zip_bytes.writestr(k, v)
        zip_bytes.close()

        # unzip
        with open("test.zip", "rb") as f:
            extracted_files = ops.unzip_obsidian(f.read())
        for filename, meta in extracted_files.md.items():
            self.assertEqual(
                self.orig_data[filename],
                meta.file,
                msg=filename)
        os.remove("test.zip")

    def test_unzip_not_zip(self):
        with open("test.zip", "wb") as f:
            f.write(b"hello world")
        with open("test.zip", "rb") as f:
            with self.assertRaises(BadZipFile):
                _ = ops.unzip_obsidian(f.read())
        os.remove("test.zip")

    def test_file_hash(self):
        for h, b in [
            ("d41d8cd98f00b204e9800998ecf8427e", b""),
            ("9e107d9d372bb6826bd81d3542a419d6", b"The quick brown fox jumps over the lazy dog")
        ]:
            file = saver.File(
                data=io.BytesIO(b),
                filename="aa.md",
            )
            self.assertEqual(".md", file.ext)
            self.assertEqual(const.app.FileTypesEnum.PLAIN, file.type)
            self.assertEqual(h + ".md", file.hashed_filename)

    async def test_replace_inner_link(self):
        md = dedent("""\
            # 123
            ddd qwd [[123]] 345
            [[我哦]]
            """)
        o1 = str(ObjectId())
        o2 = str(ObjectId())
        exist_path2nid = {"123.md": o1, "我哦.md": o2}
        res = await ops.replace_inner_link_and_upload(
            uid="",
            md=md,
            exist_path2nid=exist_path2nid,
            others_full={},
            others_name={},
        )
        self.assertEqual(dedent(f"""\
            # 123
            ddd qwd [@123](/n/{o1}) 345
            [@我哦](/n/{o2})
            """), res)
        self.assertEqual({"123.md": o1, "我哦.md": o2}, exist_path2nid)

    async def test_replace_inner_link_new(self):
        md = dedent("""\
            # 123
            ddd qwd [[123]] 345
            [[我哦]]
            """)

        o1 = str(ObjectId())
        exist_path2nid = {"123.md": o1}
        res = await ops.replace_inner_link_and_upload(
            uid="",
            md=md,
            exist_path2nid=exist_path2nid,
            others_full={},
            others_name={},
        )
        self.assertEqual(dedent(f"""\
            # 123
            ddd qwd [@123](/n/{o1}) 345
            [@我哦](/n/{exist_path2nid['我哦.md']})
            """), res)
        self.assertIn("我哦.md", exist_path2nid)



import unittest
from unittest.mock import patch

from fastapi import HTTPException

from retk.controllers import oauth
from retk.depend.sso import github
from tests import utils


class SSOTest(unittest.IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        utils.set_env(".env.test.local")
        oauth.init_oauth_provider_map()

    @classmethod
    def tearDownClass(cls) -> None:
        utils.drop_env(".env.test.local")

    async def test_github(self):
        sso = github.GithubSSO(
            client_id="",
            client_secret="",
            redirect_uri="",
            allow_insecure_http=False,
            use_state=False,
            scope=None,
        )
        self.assertEqual("github", sso.provider)
        self.assertEqual({
            "authorization_endpoint": "https://github.com/login/oauth/authorize",
            "token_endpoint": "https://github.com/login/oauth/access_token",
            "userinfo_endpoint": "https://api.github.com/user",
        }, await sso.get_discovery_document())

    async def test_github_openid_from_response(self):
        sso = github.GithubSSO(
            client_id="",
            client_secret="",
            redirect_uri="",
            allow_insecure_http=False,
            use_state=False,
            scope=None,
        )
        open_id = await sso.openid_from_response({
            "email": "email",
            "id": "id",
            "login": "login",
            "avatar_url": "avatar_url",
        })
        self.assertEqual("id", open_id.id)
        self.assertEqual("github", open_id.provider)
        self.assertEqual("login", open_id.display_name)
        self.assertEqual("avatar_url", open_id.picture)
        self.assertEqual("email", open_id.email)

    @patch("retk.depend.sso.base.SSOBase.get_login_url")
    async def test_login_github(
            self,
            mock_get_login_url,
    ):
        mock_get_login_url.return_value = "https://github.com/"
        res = await oauth.login_provider("github")
        self.assertEqual("https://github.com/", res.uri.unicode_string())

    async def test_login_not_exist(self):
        with self.assertRaises(HTTPException):
            await oauth.login_provider("not_exist")



from typing import Dict

from fastapi import Request
from fastapi.responses import JSONResponse

from retk import config, const, core, utils
from retk.controllers.account import set_cookie_response
from retk.controllers.utils import json_exception
from retk.depend.sso.base import SSOLoginError, SSOBase
from retk.depend.sso.facebook import FacebookSSO
from retk.depend.sso.github import GithubSSO
from retk.depend.sso.google import GoogleSSO
# from retk.depend.sso.qq import QQSSO
from .schemas.oauth import OAuthResponse

sso_map: Dict[str, SSOBase] = {}
user_source_map: Dict[str, int] = {}


def init_oauth_provider_map():
    sso_map.update({
        "github": GithubSSO(
            client_id=config.get_settings().OAUTH_CLIENT_ID_GITHUB,
            client_secret=config.get_settings().OAUTH_CLIENT_SEC_GITHUB,
            redirect_uri=f"{config.get_settings().OAUTH_REDIRECT_URL}/github",
            allow_insecure_http=False,
            use_state=False,
        ),
        "facebook": FacebookSSO(
            client_id=config.get_settings().OAUTH_CLIENT_ID_FACEBOOK,
            client_secret=config.get_settings().OAUTH_CLIENT_SEC_FACEBOOK,
            redirect_uri=f"{config.get_settings().OAUTH_REDIRECT_URL}/facebook",
            allow_insecure_http=False,
            use_state=False,
        ),
        "google": GoogleSSO(
            client_id=config.get_settings().OAUTH_CLIENT_ID_GOOGLE,
            client_secret=config.get_settings().OAUTH_CLIENT_SEC_GOOGLE,
            redirect_uri=f"{config.get_settings().OAUTH_REDIRECT_URL}/google",
            allow_insecure_http=False,
            use_state=False,
        ),
    })
    user_source_map.update({
        "github": const.UserSourceEnum.GITHUB.value,
        "facebook": const.UserSourceEnum.FACEBOOK.value,
        "google": const.UserSourceEnum.GOOGLE.value,
    })


async def login_provider(provider_name: str) -> OAuthResponse:
    try:
        p = sso_map[provider_name]
    except KeyError:
        raise json_exception(
            request_id="",
            code=const.CodeEnum.OAUTH_PROVIDER_NOT_FOUND,
            language=const.LanguageEnum.EN.value,
        )
    return OAuthResponse(
        uri=await p.get_login_url(),
    )


async def provider_callback(provider_name: str, req: Request) -> JSONResponse:
    try:
        p = sso_map[provider_name]
        user_source = user_source_map[provider_name]
    except KeyError:
        raise json_exception(
            request_id="",
            code=const.CodeEnum.OAUTH_PROVIDER_NOT_FOUND,
            language=const.LanguageEnum.EN.value,
        )
    try:
        user = await p.verify_and_process(req)
    except SSOLoginError:
        raise json_exception(
            request_id="",
            code=const.CodeEnum.INVALID_AUTH,
        )
    if user is None:
        raise json_exception(
            request_id="",
            code=const.CodeEnum.INVALID_AUTH,
        )
    u, code = await core.user.get_account(account=user.id, source=user_source)
    if code == const.CodeEnum.OK:
        access_token, refresh_token = utils.get_token(
            uid=u["id"],
            language=u["settings"]["language"],
        )

        return set_cookie_response(
            uid=u["id"],
            req_id="",
            status_code=200,
            access_token=access_token,
            refresh_token=refresh_token,
        )

    # no user found, create one
    language = const.LanguageEnum.EN.value
    u, code = await core.user.add(
        account=user.id,
        source=user_source,
        email=user.email if user.email else "",
        hashed="",
        nickname=user.display_name if user.display_name else "",
        avatar=user.picture if user.picture else "",
        language=language,
    )

    if code != const.CodeEnum.OK:
        raise json_exception(
            request_id="",
            code=code,
        )

    code = await core.node.new_user_add_default_nodes(language=language, uid=u["id"])
    if code != const.CodeEnum.OK:
        raise json_exception(
            request_id="",
            code=code,
        )

    access_token, refresh_token = utils.get_token(
        uid=u["id"],
        language=language,
    )
    return set_cookie_response(
        uid=u["id"],
        req_id="",
        status_code=201,
        access_token=access_token,
        refresh_token=refresh_token,
    )



from typing import Optional

from fastapi import APIRouter, Query

from retk.controllers import schemas
from retk.controllers.node import trash_ops
from retk.routes import utils

router = APIRouter(
    prefix="/api/trash",
    tags=["trash"],
    responses={404: {"description": "Not found"}},
)


@router.get(
    path="/",
    status_code=200,
    response_model=schemas.node.NodesSearchResponse,
)
@utils.measure_time_spend
async def get_from_trash(
        au: utils.ANNOTATED_AUTHED_USER,
        p: int = Query(default=0, ge=0, description="page number"),
        limit: int = Query(default=10, ge=0, le=200, description="page size"),
        referer: Optional[str] = utils.DEPENDS_REFERER,
) -> schemas.node.NodesSearchResponse:
    return await trash_ops.get_from_trash(
        au=au,
        p=p,
        limit=limit,
    )


# has to be before /{nid} otherwise it will be treated as a nid
@router.put(
    path="/batch",
    status_code=200,
    response_model=schemas.RequestIdResponse,
)
@utils.measure_time_spend
async def batch_nodes_to_trash(
        au: utils.ANNOTATED_AUTHED_USER,
        req: schemas.node.BatchNodeIdsRequest,
        referer: Optional[str] = utils.DEPENDS_REFERER,
) -> schemas.RequestIdResponse:
    return await trash_ops.move_batch_to_trash(
        au=au,
        req=req,
    )


@router.put(
    path="/batch/restore",
    status_code=200,
    response_model=schemas.RequestIdResponse,
)
@utils.measure_time_spend
async def restore_batch_node_in_trash(
        au: utils.ANNOTATED_AUTHED_USER,
        req: schemas.node.BatchNodeIdsRequest,
        referer: Optional[str] = utils.DEPENDS_REFERER,
) -> schemas.RequestIdResponse:
    return await trash_ops.restore_batch_from_trash(
        au=au,
        req=req,
    )


@router.put(
    path="/batch/delete",
    status_code=200,
    response_model=schemas.RequestIdResponse,
)
@utils.measure_time_spend
async def delete_batch_node(
        au: utils.ANNOTATED_AUTHED_USER,
        req: schemas.node.BatchNodeIdsRequest,
        referer: Optional[str] = utils.DEPENDS_REFERER,
) -> schemas.RequestIdResponse:
    return await trash_ops.delete_batch_node(
        au=au,
        req=req,
    )


@router.put(
    path="/{nid}",
    status_code=200,
    response_model=schemas.RequestIdResponse,
)
@utils.measure_time_spend
async def move_to_trash(
        au: utils.ANNOTATED_AUTHED_USER,
        nid: str = utils.ANNOTATED_NID,
        referer: Optional[str] = utils.DEPENDS_REFERER,
) -> schemas.RequestIdResponse:
    return await trash_ops.move_to_trash(
        au=au,
        nid=nid,
    )


@router.put(
    path="/{nid}/restore",
    status_code=200,
    response_model=schemas.RequestIdResponse,
)
@utils.measure_time_spend
async def restore_node_in_trash(
        au: utils.ANNOTATED_AUTHED_USER,
        nid: str = utils.ANNOTATED_NID,
        referer: Optional[str] = utils.DEPENDS_REFERER,
) -> schemas.RequestIdResponse:
    return await trash_ops.restore_from_trash(
        au=au,
        nid=nid,
    )


@router.delete(
    path="/{nid}",
    response_model=schemas.RequestIdResponse,
)
@utils.measure_time_spend
async def delete_node(
        au: utils.ANNOTATED_AUTHED_USER,
        nid: str = utils.ANNOTATED_NID,
        referer: Optional[str] = utils.DEPENDS_REFERER,
) -> schemas.RequestIdResponse:
    return await trash_ops.delete_node(
        au=au,
        nid=nid,
    )



import os
from pathlib import Path
from typing import Literal, Union

import uvicorn

from retk.core import scheduler
from retk.plugins.base import event_plugin_map


def _plugins_start_on_schedule():
    for plugin in event_plugin_map["on_schedule"]:
        if plugin.schedule_timing is None or not plugin.activated:
            continue
        kw = {}
        if plugin.schedule_timing.every == scheduler.timing.Every.minute:
            kw["second"] = plugin.schedule_timing.at_second
        elif plugin.schedule_timing.every == scheduler.timing.Every.hour:
            kw["second"] = plugin.schedule_timing.at_second
            kw["minute"] = plugin.schedule_timing.at_minute
        elif plugin.schedule_timing.every == scheduler.timing.Every.day:
            kw["second"] = plugin.schedule_timing.at_second
            kw["minute"] = plugin.schedule_timing.at_minute
            kw["hour"] = plugin.schedule_timing.at_hour
        elif plugin.schedule_timing.every == scheduler.timing.Every.month:
            kw["second"] = plugin.schedule_timing.at_second
            kw["minute"] = plugin.schedule_timing.at_minute
            kw["hour"] = plugin.schedule_timing.at_hour
            kw["day"] = plugin.schedule_timing.at_day
        scheduler.run_every_at(
            func=plugin.on_schedule,
            **kw,
            args=(plugin, True),
        )


def run(
        path: Union[str, Path] = None,
        host: str = "127.0.0.1",
        port: int = 8000,
        language: Literal["en", "zh"] = "en",
        password: str = None,
        headless: bool = False,
        debug: bool = False,
):
    """
    Run the server.

    Args:
        path (str): the database path for text and file data.
            Rethink will create .data folder in this path, default is current working directory.
        host (str): server host, default is the localhost.
        port (int): server port, default is 8000.
        language (str): website language, default is English.
            Possible values are "en" and "zh".
        password (str): the password for the server. default is None. If a string is provided,
            the login page will appear.
        headless (bool): run the server in headless mode. if False, it will open a browser after startup.
            default is False.
        debug (bool): run the server in debug mode. default is False.

    Raises:
        FileNotFoundError: if the path not exists.
        NotADirectoryError: if the path is not a directory.
    """
    if path is None:
        path = os.getcwd()
    if isinstance(path, str):
        path = Path(path)
    if not path.exists():
        raise FileNotFoundError(f"Path not exists: {path}")
    if not path.is_dir():
        raise NotADirectoryError(f"Path is not a directory: {path}")
    os.environ["VUE_APP_API_URL"] = f"http://{host}:{port}"
    os.environ["VUE_APP_MODE"] = "local"
    os.environ["RETHINK_DEFAULT_LANGUAGE"] = language
    os.environ["RETHINK_LOCAL_STORAGE_PATH"] = str(path)
    os.environ["RETHINK_SERVER_HEADLESS"] = "1" if headless else "0"
    os.environ["RETHINK_SERVER_DEBUG"] = "true" if debug else "false"
    if password is not None:
        l_pw = len(password)
        if l_pw < 6 or l_pw > 20:
            raise ValueError("Password length should be between 6 and 20 characters!")
        if len(password) < 8:
            print("Password is too short. The password is highly recommended to be at least 8 characters!")
        os.environ["RETHINK_SERVER_PASSWORD"] = password

    _plugins_start_on_schedule()
    uvicorn.run(
        "retk.application:app",
        host=host,
        port=port,
        reload=False,
        workers=1,
        log_level="error",
        env_file=os.path.join(os.path.abspath(os.path.dirname(__file__)), ".env.local"),
    )



from typing import Optional

from fastapi import APIRouter
from fastapi.params import Query

from retk.controllers import schemas, user
from retk.routes import utils

router = APIRouter(
    prefix="/api/users",
    tags=["user"],
    responses={404: {"description": "Not found"}},
)


@router.get(
    path="/",
    status_code=200,
    response_model=schemas.user.UserInfoResponse,
)
@utils.measure_time_spend
async def get_user(
        au: utils.ANNOTATED_AUTHED_USER,
        referer: Optional[str] = utils.DEPENDS_REFERER,
) -> schemas.user.UserInfoResponse:
    return await user.get_user(
        au=au,
    )


@router.patch(
    path="/",
    status_code=200,
    response_model=schemas.user.UserInfoResponse,
)
@utils.measure_time_spend
async def update_user(
        au: utils.ANNOTATED_AUTHED_USER,
        req: schemas.user.PatchUserRequest,
        referer: Optional[str] = utils.DEPENDS_REFERER,
) -> schemas.user.UserInfoResponse:
    return await user.patch_user(
        au=au,
        req=req,
    )


@router.put(
    path="/password",
    status_code=200,
    response_model=schemas.RequestIdResponse,
)
@utils.measure_time_spend
async def update_user_password(
        au: utils.ANNOTATED_AUTHED_USER,
        req: schemas.user.UpdatePasswordRequest,
        referer: Optional[str] = utils.DEPENDS_REFERER,
) -> schemas.RequestIdResponse:
    return await user.update_password(
        au=au,
        req=req,
    )


@router.get(
    path="/notices",
    status_code=200,
    response_model=schemas.user.NotificationResponse,
)
@utils.measure_time_spend
async def get_notifications(
        au: utils.ANNOTATED_AUTHED_USER,
        unread: Optional[bool] = Query(default=False),
        p: int = Query(0, ge=0),
        limit: int = Query(10, ge=1, le=100),
        referer: Optional[str] = utils.DEPENDS_REFERER,
) -> schemas.user.NotificationResponse:
    return await user.get_user_notices(
        au=au,
        unread_only=unread,
        page=p,
        limit=limit,
    )


@router.put(
    path="/notices/system/read/{notice_id}",
    status_code=200,
    response_model=schemas.RequestIdResponse,
)
@utils.measure_time_spend
async def mark_read_notice(
        au: utils.ANNOTATED_AUTHED_USER,
        notice_id: str,
        referer: Optional[str] = utils.DEPENDS_REFERER,
) -> schemas.RequestIdResponse:
    return await user.mark_system_notice_read(
        au=au,
        notice_id=notice_id,
    )


@router.put(
    path="/notices/system/read-all",
    status_code=200,
    response_model=schemas.RequestIdResponse,
)
@utils.measure_time_spend
async def mark_read_all_notices(
        au: utils.ANNOTATED_AUTHED_USER,
        referer: Optional[str] = utils.DEPENDS_REFERER,
) -> schemas.RequestIdResponse:
    return await user.mark_all_system_notice_read(
        au=au,
    )



"""Facebook SSO Login Helper
"""

from .base import DiscoveryDocument, OpenID, SSOBase


class FacebookSSO(SSOBase):
    """Class providing login via Facebook OAuth"""

    provider = "facebook"
    base_url = "https://graph.facebook.com/v9.0"
    scope = ["email"]

    async def get_discovery_document(self) -> DiscoveryDocument:
        """Get document containing handy urls"""
        return {
            "authorization_endpoint": "https://www.facebook.com/v9.0/dialog/oauth",
            "token_endpoint": f"{self.base_url}/oauth/access_token",
            "userinfo_endpoint": f"{self.base_url}/me?fields=id,name,email,first_name,last_name,picture",
        }

    @classmethod
    async def openid_from_response(cls, response: dict) -> OpenID:
        """Return OpenID from user information provided by Facebook"""
        return OpenID(
            email=response.get("email", ""),
            first_name=response.get("first_name"),
            last_name=response.get("last_name"),
            display_name=response.get("name"),
            provider=cls.provider,
            id=response.get("id"),
            picture=response.get("picture", {}).get("data", {}).get("url", None),
        )



import os
from contextlib import asynccontextmanager

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware

from retk import const, config, safety, utils
from retk.controllers.oauth import init_oauth_provider_map
from retk.controllers.self_hosted import notice_new_pkg_version
from retk.core import scheduler
from retk.core.files.importing import async_tasks
from retk.logger import logger, add_rotating_file_handler
from retk.plugins.register import register_official_plugins
from .models.client import client
from .routes import (
    user,
    oauth,
    node,
    recent,
    trash,
    app_captcha,
    files,
    plugin,
    self_hosted,
    account,
    manager,
    statistic,
    notice,
)


@asynccontextmanager
async def lifespan(app: FastAPI):
    # on startup
    if not config.is_local_db():
        add_rotating_file_handler(
            log_dir=const.settings.RETHINK_DIR.parent.parent / "logs",
            max_bytes=10 * 1024 * 1024,
            backup_count=10,
        )
    logger.debug(f'startup_event RETHINK_LOCAL_STORAGE_PATH: {os.environ.get("RETHINK_LOCAL_STORAGE_PATH")}')
    logger.debug(f'startup_event VUE_APP_MODE: {os.environ.get("VUE_APP_MODE")}')
    logger.debug(f'startup_event VUE_APP_API_URL: {os.environ.get("VUE_APP_API_URL")}')
    logger.debug(f'startup_event RETHINK_DEFAULT_LANGUAGE: {os.environ.get("RETHINK_DEFAULT_LANGUAGE")}')
    await client.init()

    # schedule job
    scheduler.start()
    scheduler.init_tasks()

    # init oauth provider map
    init_oauth_provider_map()

    # notice new pkg version
    if config.is_local_db():
        await notice_new_pkg_version()

    # register official plugins
    register_official_plugins()

    # local finish up
    utils.local_finish_up()

    yield

    # on shutdown
    scheduler.stop()
    await client.close()
    await client.search.close()
    logger.debug("fastapi shutdown event: db and searcher closed")

    async_tasks.stop()
    logger.debug("fastapi shutdown event: async_tasks stopped")


app = FastAPI(
    docs_url="/docs",
    openapi_url="/openapi.json",
    lifespan=lifespan,
)

app.add_middleware(
    CORSMiddleware,
    allow_origins=safety.allow_origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
    expose_headers=["X-Captcha-Token"],
)
app.add_middleware(safety.CSPMiddleware)
app.add_middleware(safety.FrameOptionsMiddleware)

for r in [
    user,
    oauth,
    node,
    recent,
    trash,
    app_captcha,
    files,
    plugin,
    account,
    manager,
    statistic,
    notice,
]:
    app.include_router(r.router)

# self-hosted
app.include_router(self_hosted.r_router)
app.include_router(self_hosted.node_file_router)
app.include_router(self_hosted.home_router)
self_hosted.mount_static(app=app)



from typing import Tuple, Optional

import httpx

from retk import const
from retk.logger import logger
from retk.utils import ASYNC_CLIENT_HEADERS


def parse_version(version: str) -> Optional[Tuple[int, int, int]]:
    vs = version.split(".")
    if len(vs) != 3:
        return None
    try:
        vs = (int(vs[0]), int(vs[1]), int(vs[2]))
    except ValueError:
        return None
    return vs


async def get_latest_pkg_version() -> Tuple[Tuple[int, int, int], const.CodeEnum]:
    url = 'https://pypi.org/pypi/retk/json'
    default_version = (0, 0, 0)
    async with httpx.AsyncClient() as ac:
        try:
            response = await ac.get(
                url=url,
                headers=ASYNC_CLIENT_HEADERS,
                follow_redirects=False,
                timeout=2.
            )
        except (
                httpx.ConnectTimeout,
                RuntimeError,
                httpx.ConnectError,
                httpx.ReadTimeout,
                httpx.HTTPError
        ) as e:
            logger.debug(f"failed to get {url}: {e}")
            return default_version, const.CodeEnum.OPERATION_FAILED

    if response.status_code != 200:
        logger.debug(f"failed to get {url}: {response.status_code}, {response.text}")
        return default_version, const.CodeEnum.OPERATION_FAILED

    package_info = response.json()

    try:
        v = package_info['info']['version']
    except KeyError:
        logger.debug(f"failed to get {url}: {response.text}")
        return default_version, const.CodeEnum.OPERATION_FAILED
    vs = parse_version(v)
    if vs is None:
        logger.debug(f"failed to get {url}: {v}")
        return default_version, const.CodeEnum.OPERATION_FAILED
    return vs, const.CodeEnum.OK



import time
import traceback
from functools import wraps
from typing import Optional

import jwt
from fastapi import HTTPException, Header, Cookie, Depends
from fastapi.params import Path
from starlette.status import HTTP_403_FORBIDDEN
from typing_extensions import Annotated

from retk import const, config, core
from retk.controllers.utils import json_exception
from retk.logger import logger
from retk.models.tps import AuthedUser, convert_user_dict_to_authed_user
from retk.utils import jwt_decode

REFERER_PREFIX = f"https://{const.settings.DOMAIN}"


def measure_time_spend(func):
    @wraps(func)
    async def wrapper(*args, **kwargs):
        t0 = time.perf_counter()
        uid = ""
        try:
            au: AuthedUser = kwargs["au"]
        except KeyError:
            req_id = ""
        else:
            req_id = au.request_id
            if au.u is not None:
                uid = au.u.id

        req_s = str(kwargs.get("req", ""))

        req_s = req_s[:200] + "..." if len(req_s) > 200 else req_s
        if func.__name__ not in ["login", "forget_password", "signup"]:
            logger.debug(f"REQ: reqId='{req_id}' | uid='{uid}' | api='{func.__name__}' | {req_s}")

        resp = await func(*args, **kwargs)
        t1 = time.perf_counter()
        try:
            req_id = resp.requestId
        except AttributeError:
            req_id = ""
        logger.debug(f"RESP: reqId='{req_id}' | uid='{uid}' | api='{func.__name__}' | spend={t1 - t0:.4f}s")
        return resp

    return wrapper


def verify_referer(referer: Optional[str] = Header(None)):
    if config.get_settings().VERIFY_REFERER and not referer.startswith(REFERER_PREFIX):
        logger.error(f"referer={referer} not startswith {REFERER_PREFIX}")
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN,
            detail="Invalid referer",
        )
    return referer


async def __process_auth_headers(  # noqa: C901
        is_refresh_token: bool,
        refresh_token_id: str,
        token: str = Header(alias="Authorization", default=""),
        request_id: str = Header(
            default="", alias="RequestId", max_length=const.settings.MD_MAX_LENGTH
        )
) -> AuthedUser:
    if token is None or token == "":
        raise json_exception(
            request_id=request_id,
            code=const.CodeEnum.INVALID_AUTH if is_refresh_token else const.CodeEnum.EXPIRED_OR_NO_ACCESS_TOKEN,
            log_msg="empty token",
        )
    au = AuthedUser(
        u=None,
        request_id=request_id,
        language=const.LanguageEnum.EN.value,
    )
    err = ""
    u = None
    try:
        payload = jwt_decode(token=token)
        is_access = payload.get("is_access")
        if is_access is None:
            code = const.CodeEnum.INVALID_AUTH
            err = "invalid token"
        elif is_access == is_refresh_token:
            code = const.CodeEnum.INVALID_AUTH
            err = "invalid token"
        elif is_refresh_token and payload["uid"] != refresh_token_id:
            code = const.CodeEnum.INVALID_AUTH
            err = "invalid token"
        else:
            u, code = await core.user.get(uid=payload["uid"], disabled=False)
            if code != const.CodeEnum.OK:
                err = f"get user failed, code={code}"

    except jwt.exceptions.ExpiredSignatureError:
        code = const.CodeEnum.EXPIRED_AUTH if is_refresh_token else const.CodeEnum.EXPIRED_OR_NO_ACCESS_TOKEN
        err = "auth expired"
    except jwt.exceptions.DecodeError:
        code = const.CodeEnum.INVALID_AUTH
        err = "token decode error"
    except jwt.exceptions.InvalidTokenError:
        code = const.CodeEnum.INVALID_AUTH
        err = "invalid token"
    except Exception:  # pylint: disable=broad-except
        code = const.CodeEnum.INVALID_AUTH
        err = traceback.format_exc()
    if code != const.CodeEnum.OK or u is None:
        raise json_exception(
            request_id=request_id,
            code=code,
            log_msg=err,
        )
    au.language = u["settings"]["language"]
    au.u = convert_user_dict_to_authed_user(u)
    return au


async def process_normal_headers(
        token: str = Cookie(alias=const.settings.COOKIE_ACCESS_TOKEN, default=""),
        request_id: str = Header(
            default="", alias="RequestId", max_length=const.settings.MD_MAX_LENGTH
        )
) -> AuthedUser:
    return await __process_auth_headers(
        is_refresh_token=False,
        refresh_token_id="",
        token=token,
        request_id=request_id
    )


async def process_refresh_token_headers(
        token: str = Cookie(alias=const.settings.COOKIE_REFRESH_TOKEN, default=""),
        id_: str = Cookie(alias=const.settings.COOKIE_REFRESH_TOKEN_ID, default=""),
        request_id: str = Header(
            default="", alias="RequestId", max_length=const.settings.MD_MAX_LENGTH
        )
) -> AuthedUser:
    return await __process_auth_headers(
        is_refresh_token=True,
        refresh_token_id=id_,
        token=token,
        request_id=request_id
    )


async def process_no_auth_headers(
        request_id: str = Header(
            default="", alias="RequestId", max_length=const.settings.MD_MAX_LENGTH
        )
) -> str:
    return request_id


async def process_admin_headers(
        token: str = Cookie(alias=const.settings.COOKIE_ACCESS_TOKEN, default=""),
        request_id: str = Header(
            default="", alias="RequestId", max_length=const.settings.MD_MAX_LENGTH
        )
) -> AuthedUser:
    au = await process_normal_headers(token=token, request_id=request_id)
    if au.u is None or au.u.type != const.USER_TYPE.ADMIN.id:
        raise json_exception(
            request_id=au.request_id,
            code=const.CodeEnum.NOT_PERMITTED,
        )
    return au


ANNOTATED_REQUEST_ID = Annotated[str, Depends(process_no_auth_headers)]
ANNOTATED_AUTHED_USER = Annotated[AuthedUser, Depends(process_normal_headers)]
ANNOTATED_AUTHED_ADMIN = Annotated[AuthedUser, Depends(process_admin_headers)]
ANNOTATED_REFRESH_TOKEN = Annotated[AuthedUser, Depends(process_refresh_token_headers)]

ANNOTATED_UID = Annotated[str, Path(title="The ID of user", max_length=const.settings.UID_MAX_LENGTH, min_length=8)]
ANNOTATED_NID = Annotated[str, Path(title="The ID of node", max_length=const.settings.NID_MAX_LENGTH, min_length=8)]
ANNOTATED_PID = Annotated[str, Path(title="The ID of plugin", max_length=const.settings.PLUGIN_ID_MAX_LENGTH)]
ANNOTATED_FID = Annotated[str, Path(title="The ID of file", max_length=const.settings.FID_MAX_LENGTH)]

DEPENDS_REFERER = Depends(verify_referer)



from datetime import datetime
from typing import List, Optional, Tuple, TypedDict

from bson import ObjectId
from bson.tz_util import utc

from retk import const, config
from retk.models.client import client
from retk.models.tps import AuthedUser, NoticeManagerDelivery
from retk.utils import datetime2str, md2html, md2txt


async def post_in_manager_delivery(
        au: AuthedUser,
        title: str,
        content: str,
        recipient_type: int,
        batch_type_ids: List[str],
        publish_at: Optional[datetime] = None  # if None, use current
) -> Tuple[Optional[NoticeManagerDelivery], const.CodeEnum]:
    if const.USER_TYPE.id2config(au.u.type) not in (const.USER_TYPE.ADMIN, const.USER_TYPE.MANAGER):
        return None, const.CodeEnum.NOT_PERMITTED
    if publish_at is None:
        publish_at = datetime.now(tz=utc)

    # if publish_at is not utc, convert it to utc
    if publish_at.tzinfo is None or publish_at.tzinfo.utcoffset(publish_at) is None:
        publish_at = publish_at.replace(tzinfo=utc)
    elif publish_at.tzinfo != utc:
        publish_at = publish_at.astimezone(utc)

    # add system notice
    notice: NoticeManagerDelivery = {
        "_id": ObjectId(),
        "senderType": au.u.type,
        "senderId": au.u.id,
        "title": title,
        "html": md2html(content),
        "snippet": md2txt(content)[:20],
        "recipientType": recipient_type,  # send to which user type, 0: all, 1: batch, 2: admin, 3: manager
        "batchTypeIds": batch_type_ids,  # if recipient=batch, put user id here
        "publishAt": publish_at,  # publish time
        "scheduled": False,  # has been scheduled to sent to user
    }
    res = await client.coll.notice_manager_delivery.insert_one(notice)
    if not res.acknowledged:
        return None, const.CodeEnum.OPERATION_FAILED
    return notice, const.CodeEnum.OK


async def get_system_notices(
        page: int,
        limit: int,
) -> Tuple[List[NoticeManagerDelivery], int]:
    total = await client.coll.notice_manager_delivery.count_documents({})
    notices = await client.coll.notice_manager_delivery.find(
        {}
    ).skip(page * limit).limit(limit=limit).to_list(None)
    return notices, total


async def get_system_notice(
        uid: str,
        notice_id: str,
) -> Tuple[Optional[NoticeManagerDelivery], const.CodeEnum]:
    notice = await client.coll.notice_manager_delivery.find_one(
        {"_id": ObjectId(notice_id), "scheduled": True}
    )
    if notice is None or notice["senderType"] not in [const.USER_TYPE.ADMIN.id, const.USER_TYPE.MANAGER.id]:
        # check if the notice exists and is sent by admin or manager
        return None, const.CodeEnum.NOTICE_NOT_FOUND
    if notice["recipientType"] == const.notice.RecipientTypeEnum.ALL:
        # this notice is for all users
        return notice, const.CodeEnum.OK
    elif notice["recipientType"] == const.notice.RecipientTypeEnum.BATCH:
        # this notice is for some users
        if uid in notice["batchTypeIds"]:
            # check if the user is in the batch
            return notice, const.CodeEnum.OK
        return None, const.CodeEnum.NOTICE_NOT_FOUND
    else:
        # this notice is for a specific user type
        u = await client.coll.users.find_one({"id": uid})
        if u is None or u["type"] != notice["recipientType"]:
            # check if the user is the recipient type
            return None, const.CodeEnum.NOTICE_NOT_FOUND
    return notice, const.CodeEnum.OK


class Notice(TypedDict):
    id: str
    title: str
    snippet: str
    publishAt: str
    read: bool
    readTime: Optional[str]


class SystemNotices(TypedDict):
    total: int
    notices: List[Notice]


class Notices(TypedDict):
    hasUnread: bool
    system: SystemNotices


async def get_user_notices(
        au: AuthedUser,
        unread_only: bool = False,
        page: int = 0,
        limit: int = 10,
) -> Tuple[Notices, const.CodeEnum]:
    c = {"recipientId": au.u.id}
    if unread_only:
        c["read"] = False
    if not config.is_local_db():
        user_system_notices = await client.coll.notice_system.find(
            c,
            projection={"noticeId": 1, "read": 1, "readTime": 1}
        ).sort("_id", -1).skip(page * limit).limit(limit=limit).to_list(None)
        # Get the details of the notices
        n_details = await client.coll.notice_manager_delivery.find(
            {"_id": {"$in": [n["noticeId"] for n in user_system_notices]}},
            projection={"title": 1, "snippet": 1, "publishAt": 1}
        ).to_list(None)
    else:
        user_system_notices = await client.coll.notice_system.find(
            c,
        ).sort("_id", -1).skip(page * limit).limit(limit=limit).to_list(None)
        user_system_notices = [{
            "noticeId": n["noticeId"],
            "read": n["read"],
            "readTime": n["readTime"],
        } for n in user_system_notices]
        # Get the details of the notices
        n_details = await client.coll.notice_manager_delivery.find(
            {"_id": {"$in": [n["noticeId"] for n in user_system_notices]}},
        ).to_list(None)

    total_system_system = await client.coll.notice_system.count_documents(c)
    if c.get("read", True):
        # if not unread_only, check if there are unread notices
        has_unread = await client.coll.notice_system.count_documents({"recipientId": au.u.id, "read": False}) > 0
    else:
        has_unread = total_system_system > 0

    n_details_dict = {n["_id"]: n for n in n_details}
    new_system_notices: List[Notice] = []
    for usn in user_system_notices:
        detail = n_details_dict[usn["noticeId"]]
        new_system_notices.append({
            "id": str(usn["noticeId"]),
            "title": detail["title"],
            "snippet": detail["snippet"],
            "publishAt": datetime2str(detail["publishAt"]),
            "read": usn["read"],
            "readTime": datetime2str(usn["readTime"]) if usn["readTime"] is not None else None,
        })

    return {
        "hasUnread": has_unread,
        "system": {
            "total": total_system_system,
            "notices": new_system_notices,
        }
    }, const.CodeEnum.OK


async def mark_system_notice_read(
        uid: str,
        notice_id: str,
) -> const.CodeEnum:
    res = await client.coll.notice_system.update_one(
        {"recipientId": uid, "noticeId": ObjectId(notice_id)},
        {"$set": {"read": True, "readTime": datetime.now(tz=utc)}}
    )
    if not res.acknowledged:
        return const.CodeEnum.OPERATION_FAILED
    return const.CodeEnum.OK


async def mark_all_system_notice_read(
        au: AuthedUser,
) -> const.CodeEnum:
    res = await client.coll.notice_system.update_many(
        {"recipientId": au.u.id},
        {"$set": {"read": True, "readTime": datetime.now(tz=utc)}}
    )
    if not res.acknowledged:
        return const.CodeEnum.OPERATION_FAILED
    return const.CodeEnum.OK



from datetime import timedelta
from io import BytesIO
from random import choices
from typing import Tuple, Dict

import jwt
from captcha.audio import AudioCaptcha
from captcha.image import ImageCaptcha

from retk import const, config
from retk.utils import jwt_encode, jwt_decode

DEFAULT_CAPTCHA_EXPIRE_SECOND = 60

img_captcha = ImageCaptcha(font_sizes=(35, 30, 32))
audio_captcha = AudioCaptcha()

alphabet = "347ACEFGJLMNPRTY"
alphabet_len = len(alphabet)
code_idx_range = list(range(0, alphabet_len - 1))


def generate(length: int = 4, sound: bool = False) -> Tuple[str, Dict[str, BytesIO]]:
    code = [alphabet[i] for i in choices(code_idx_range, k=length)]
    code_str = "".join(code)
    data = {
        "img": img_captcha.generate(code_str),
    }
    if sound:
        data["sound"] = audio_captcha.generate(code_str)
    token = jwt_encode(
        exp_delta=timedelta(seconds=DEFAULT_CAPTCHA_EXPIRE_SECOND),
        data={"code": code_str.lower() + config.get_settings().CAPTCHA_SALT}
    )
    # TODO: 以后有 redis 之后要考虑 token 的一次性校验。校验过一次后，需要主动失效
    return token, data


def verify_captcha(token: str, code_str: str) -> const.CodeEnum:
    code = const.CodeEnum.CAPTCHA_ERROR
    try:
        data = jwt_decode(token)
        if data["code"] == code_str.lower() + config.get_settings().CAPTCHA_SALT:
            code = const.CodeEnum.OK
    except jwt.ExpiredSignatureError:
        code = const.CodeEnum.CAPTCHA_EXPIRED
    except (jwt.DecodeError, Exception):  # pylint: disable=broad-except
        code = const.CodeEnum.INVALID_AUTH
    return code



import datetime
import html
from typing import Optional, Tuple

from bson import ObjectId
from bson.tz_util import utc
from pymongo.errors import DuplicateKeyError

from retk import const, config, utils
from retk.controllers.schemas.user import PatchUserRequest
from retk.models import tps
from retk.models.client import client


async def add(
        account: str,
        source: int,
        email: str,
        hashed: str,
        nickname: str,
        avatar: str,
        language: str,
) -> Tuple[Optional[tps.UserMeta], const.CodeEnum]:
    if await client.coll.users.find_one({"account": account, "source": source}) is not None:
        return None, const.CodeEnum.EMAIL_OCCUPIED
    oid = ObjectId()
    # assert language in const.Language
    if not const.LanguageEnum.is_valid(language):
        language = const.LanguageEnum.EN.value
    data = utils.get_user_dict(
        _id=oid,
        uid=utils.short_uuid(),
        source=source,
        account=account,
        nickname=nickname,
        email=email,
        avatar=str(avatar),
        hashed=hashed,
        disabled=False,
        modified_at=oid.generation_time,
        used_space=0,
        type_=const.USER_TYPE.NORMAL.id,

        last_state_recent_cursor_search_selected_nids=[],
        last_state_recent_search=[],
        last_state_node_display_method=const.NodeDisplayMethodEnum.CARD.value,
        last_state_node_display_sort_key="modifiedAt",

        settings_language=language,
        settings_theme=const.app.AppThemeEnum.LIGHT.value,
        settings_editor_mode=const.app.EditorModeEnum.WYSIWYG.value,
        settings_editor_font_size=15,
        settings_editor_code_theme=const.app.EditorCodeThemeEnum.GITHUB.value,
        settings_editor_sep_right_width=200,
        settings_editor_side_current_tool_id="",
    )

    # catch if id is duplicated
    while True:
        try:
            res = await client.coll.users.insert_one(data)
            if not res.acknowledged:
                return None, const.CodeEnum.OPERATION_FAILED
            break
        except DuplicateKeyError:
            data["id"] = utils.short_uuid()
            continue

    return data, const.CodeEnum.OK


async def patch(  # noqa: C901
        au: tps.AuthedUser,
        req: PatchUserRequest,
) -> Tuple[Optional[tps.UserMeta], const.CodeEnum]:
    new_data = {}

    if req.nickname is not None:
        nickname = req.nickname.strip()
        nickname = html.escape(nickname)
        if nickname != "":
            new_data["nickname"] = nickname

    if req.avatar is not None:
        avatar = str(req.avatar).strip()
        if avatar != "":
            new_data["avatar"] = avatar

    if req.lastState is not None:
        if req.lastState.nodeDisplayMethod is not None:
            if req.lastState.nodeDisplayMethod >= len(
                    const.NodeDisplayMethodEnum) or req.lastState.nodeDisplayMethod < 0:
                return None, const.CodeEnum.INVALID_NODE_DISPLAY_METHOD
            new_data["lastState.nodeDisplayMethod"] = req.lastState.nodeDisplayMethod

        if req.lastState.nodeDisplaySortKey is not None:
            new_data["lastState.nodeDisplaySortKey"] = req.lastState.nodeDisplaySortKey

    if req.settings is not None:
        if req.settings.language is not None:
            new_data["settings.language"] = req.settings.language

        if req.settings.theme is not None:
            new_data["settings.theme"] = req.settings.theme

        if req.settings.editorMode is not None:
            new_data["settings.editorMode"] = req.settings.editorMode

        if req.settings.editorFontSize is not None:
            if not const.app.EditorFontSizeEnum.is_valid(req.settings.editorFontSize):
                return None, const.CodeEnum.INVALID_SETTING
            new_data["settings.editorFontSize"] = req.settings.editorFontSize

        if req.settings.editorCodeTheme is not None:
            new_data["settings.editorCodeTheme"] = req.settings.editorCodeTheme

        if req.settings.editorSepRightWidth is not None:
            new_data["settings.editorSepRightWidth"] = req.settings.editorSepRightWidth

        if req.settings.editorSideCurrentToolId is not None:
            new_data["settings.editorSideCurrentToolId"] = req.settings.editorSideCurrentToolId

    # has data to update
    if len(new_data) != 0:
        new_data["modifiedAt"] = datetime.datetime.now(tz=utc)

        res = await client.coll.users.update_one(
            {"id": au.u.id},
            {"$set": new_data},
        )
        if res.modified_count != 1:
            return None, const.CodeEnum.OPERATION_FAILED

    return await get(uid=au.u.id, disabled=None)


def __get_condition_filter_manager(condition: dict, exclude_manager: bool) -> dict:
    if exclude_manager:
        condition["type"] = {"$nin": [
            const.user_types.USER_TYPE.ADMIN.id,
            const.user_types.USER_TYPE.MANAGER.id,
        ]}
    return condition


async def get_by_email(
        email: str,
        disabled: Optional[bool] = False,
        exclude_manager: bool = False,
) -> Tuple[Optional[tps.UserMeta], const.CodeEnum]:
    if config.get_settings().ONE_USER:
        source = const.UserSourceEnum.LOCAL.value
    else:
        source = const.UserSourceEnum.EMAIL.value
    return await get_account(account=email, source=source, disabled=disabled, exclude_manager=exclude_manager)


async def get_account(
        account: str,
        source: int,
        disabled: Optional[bool] = False,
        exclude_manager: bool = False,
) -> Tuple[Optional[tps.UserMeta], const.CodeEnum]:
    c = {"source": source, "account": account}
    if disabled is not None:
        c["disabled"] = disabled
    c = __get_condition_filter_manager(condition=c, exclude_manager=False)  # exclude_manager)
    u = await client.coll.users.find_one(c)
    if u is None:
        return None, const.CodeEnum.ACCOUNT_OR_PASSWORD_ERROR
    return u, const.CodeEnum.OK


async def get_google_account_by_email(
        email: str,
        disabled: Optional[bool] = False,
        exclude_manager: bool = False,
) -> Tuple[Optional[tps.UserMeta], const.CodeEnum]:
    c = {
        "source": const.UserSourceEnum.GOOGLE.value,
        "email": email,
    }
    if disabled is not None:
        c["disabled"] = disabled
    c = __get_condition_filter_manager(condition=c, exclude_manager=exclude_manager)
    u = await client.coll.users.find_one(c)
    if u is None:
        return None, const.CodeEnum.ACCOUNT_OR_PASSWORD_ERROR
    return u, const.CodeEnum.OK


async def get(
        uid: str,
        disabled: Optional[bool] = False,
        exclude_manager: bool = False
) -> Tuple[Optional[tps.UserMeta], const.CodeEnum]:
    c = {"id": uid}
    if disabled is not None:
        c["disabled"] = disabled
    c = __get_condition_filter_manager(condition=c, exclude_manager=exclude_manager)
    u = await client.coll.users.find_one(c)
    if u is None:
        return None, const.CodeEnum.USER_NOT_EXIST
    if u["usedSpace"] < 0:
        # reset usedSpace to 0
        await client.coll.users.update_one(
            {"id": uid},
            {"$set": {"usedSpace": 0}}
        )
        u["usedSpace"] = 0

    return u, const.CodeEnum.OK


async def update_used_space(uid: str, delta: int) -> const.CodeEnum:
    if delta == 0:
        return const.CodeEnum.OK
    res = await client.coll.users.update_one(
        {"id": uid},
        {"$inc": {"usedSpace": delta}}
    )
    return const.CodeEnum.OK if res.modified_count == 1 else const.CodeEnum.OPERATION_FAILED


async def user_space_not_enough(au: tps.AuthedUser) -> bool:
    if config.is_local_db():
        return False
    return au.u.used_space > const.USER_TYPE.id2config(au.u.type).max_store_space


async def reset_password(uid: str, hashed: str) -> const.CodeEnum:
    res = await client.coll.users.update_one(
        {"id": uid},
        {"$set": {"hashed": hashed}}
    )
    return const.CodeEnum.OK if res.acknowledged == 1 else const.CodeEnum.OPERATION_FAILED



class InsertOneResult():
    def __init__(self, inserted_id):
        self.acknowledged = True
        self.inserted_id = inserted_id


class InsertManyResult():
    def __init__(self, documents):
        self.acknowledged = True
        self.inserted_ids = [d['_id'] for d in documents]


class UpdateResult():
    def __init__(self, matched_count, modified_count, upserted_id=None):
        self.acknowledged = True
        self.matched_count = matched_count
        self.modified_count = modified_count
        self.upserted_id = upserted_id


class DeleteResult():
    def __init__(self, deleted_count):
        self.acknowledged = True
        self.deleted_count = deleted_count



from .errors import MongitaNotImplementedError


class ReadConcern():
    def __init__(self, level=None):
        if level is not None:
            raise MongitaNotImplementedError("Mongita's ReadConcern is a dummy "
                                             "doesn't support parameters")
        self.level = None

    @property
    def document(self):
        return {}



from retk import core
from retk.controllers import schemas
from retk.controllers.utils import maybe_raise_json_exception
from retk.models.tps import AuthedUser


async def add_recent_at_node(
        au: AuthedUser,
        req: schemas.recent.AtNodeRequest
) -> schemas.RequestIdResponse:
    code = await core.recent.added_at_node(au=au, nid=req.nid, to_nid=req.toNid)
    maybe_raise_json_exception(au=au, code=code)

    return schemas.RequestIdResponse(
        requestId=au.request_id,
    )


async def get_recent_searched(
        au: AuthedUser,
) -> schemas.recent.GetRecentSearchResponse:
    return schemas.recent.GetRecentSearchResponse(
        requestId=au.request_id,
        queries=au.u.last_state.recent_search,
    )



import time
import zipfile

import pymongo.errors

from retk import const, core
from retk.logger import logger
from retk.models.tps import AuthedUser, convert_user_dict_to_authed_user
from . import ops
from .. import utils


async def upload_obsidian_task(  # noqa: C901
        bytes_data: bytes,
        filename: str,
        max_file_size: int,
        uid: str,
        request_id: str,
) -> None:
    type_ = "obsidian"
    await utils.import_set_modules()

    doc, finished = await utils.check_last_task_finished(uid=uid, type_=type_)
    if not finished:
        return

    t0 = time.time()
    try:
        unzipped_files = ops.unzip_obsidian(bytes_data)
    except zipfile.BadZipFile as e:
        await utils.set_running_false(
            uid,
            const.CodeEnum.INVALID_FILE_TYPE,
            msg=f"unzip failed: {e}"
        )
        logger.error(f"invalid file type: {filename}, uid: {uid}")
        return
    t1 = time.time()
    logger.debug(f"obsidian upload, uid={uid}, unzip time: {t1 - t0:.2f}")

    existed_path2nid = doc.get("obsidian", {}).copy()
    md_count = len(unzipped_files.md_full)
    if md_count == 0:
        await utils.set_running_false(
            uid,
            const.CodeEnum.INVALID_FILE_TYPE,
            msg="no md file found",
        )
        logger.debug(f"no md file found, uid: {uid}")
        return
    elif md_count > 2000:
        await utils.set_running_false(
            uid,
            const.CodeEnum.TOO_MANY_FILES,
            msg=f"md file count: {md_count} > 2000",
        )
        logger.debug(f"too many md files: {md_count}, uid: {uid}")
        return

    # check file size
    for full in [unzipped_files.md_full, unzipped_files.others_full]:
        for full_path, meta in full.items():
            meta: ops.UnzipObsidian.Meta

            if meta.size > max_file_size:
                await utils.set_running_false(
                    uid,
                    const.CodeEnum.TOO_LARGE_FILE,
                    msg=f"file size > {max_file_size}: {full_path}",
                )
                logger.debug(f"too large file: {full_path}, uid: {uid}")
                return

    t2 = time.time()
    logger.debug(f"obsidian upload, uid={uid}, filter time: {t2 - t1:.2f}")

    # add new md files with only title
    u, code = await core.user.get(uid=uid, disabled=False)
    au = AuthedUser(
        u=convert_user_dict_to_authed_user(u),
        request_id=request_id,
        language=u["settings"].get("language", const.LanguageEnum.EN.value),
    )
    for i, (full_path, meta) in enumerate(unzipped_files.md_full.items()):
        meta: ops.UnzipObsidian.Meta

        if full_path in existed_path2nid:
            continue

        try:
            n, code = await core.node.post(
                au=au,
                md=meta.title,
                type_=const.NodeTypeEnum.MARKDOWN.value,
            )
        except pymongo.errors.DuplicateKeyError:
            logger.error(f"duplicate key: {full_path}, uid: {uid}")
            continue
        if code != const.CodeEnum.OK:
            await utils.set_running_false(
                uid,
                code,
                msg=f"new file insert failed: {full_path}",
            )
            logger.error(f"error: {code}, filepath: {full_path}, uid: {uid}")
            return

        # add full path and short name to existed_path2nid
        existed_path2nid[full_path] = n["id"]
        if meta.filename not in existed_path2nid:
            existed_path2nid[meta.filename] = n["id"]
        if i % 20 == 0:
            doc, code = await utils.update_process(uid=uid, type_=type_, process=int(i / md_count * 10))
            if code != const.CodeEnum.OK:
                await utils.set_running_false(
                    uid,
                    code,
                    msg="process updating failed",
                )
                logger.error(f"error: {code}, filepath: {full_path}, uid: {uid}")
                return
            if not doc["running"]:
                break
    t3 = time.time()
    logger.debug(f"obsidian upload, uid={uid}, add new md time: {t3 - t2:.2f}")

    # update all md content and update md files
    for i, (full_path, meta) in enumerate(unzipped_files.md_full.items()):
        meta: ops.UnzipObsidian.Meta

        try:
            md = meta.file.decode("utf-8")
        except (FileNotFoundError, OSError, UnicodeDecodeError) as e:
            logger.error(f"error: {e}. filepath: {full_path}")
            await utils.set_running_false(
                uid,
                const.CodeEnum.FILE_OPEN_ERROR,
                msg=f"file decode utf-8 failed, {e}: {full_path}",
            )
            logger.error(f"error: {const.CodeEnum.FILE_OPEN_ERROR}, filepath: {full_path}, uid: {uid}")
            return

        md = await ops.replace_inner_link_and_upload(
            uid=uid,
            md=md,
            exist_path2nid=existed_path2nid,
            others_full=unzipped_files.others_full,
            others_name=unzipped_files.others,
        )
        md = meta.title + "\n\n" + md
        nid = existed_path2nid[full_path]
        n, _, code = await core.node.update_md(
            au=au,
            nid=nid,
            md=md,
            refresh_on_same_md=True,
        )
        if code == const.CodeEnum.NODE_NOT_EXIST:
            n, code = await core.node.post(
                au=au,
                md=md,
                type_=const.NodeTypeEnum.MARKDOWN.value,
            )
            if code != const.CodeEnum.OK:
                await utils.set_running_false(
                    uid,
                    code,
                    msg=f"file insert failed: {full_path}",
                )
                logger.error(f"error: {code}, filepath: {full_path}, uid: {uid}")
                return
            existed_path2nid[full_path] = n["id"]
        elif code != const.CodeEnum.OK:
            await utils.set_running_false(
                uid,
                code,
                msg=f"file updating failed: {full_path}",
            )
            logger.error(f"error: {code}, filepath: {full_path}, uid: {uid}")
            return
        if i % 20 == 0:
            doc, code = await utils.update_process(uid=uid, type_=type_, process=int(i / md_count * 80 + 10))
            if code != const.CodeEnum.OK:
                await utils.set_running_false(
                    uid,
                    code,
                    msg="uploading process update failed",
                )
                logger.error(f"error: {code}, filepath: {full_path}, uid: {uid}")
                return
            if not doc["running"]:
                break
    t4 = time.time()
    logger.debug(f"obsidian upload, uid={uid}, update all files time: {t4 - t3:.2f}")

    # update for old obsidian files
    count = 0
    for base_name, nid in doc["obsidian"].items():
        if base_name not in existed_path2nid:
            n, code = await core.node.get(au=au, nid=nid)
            if code != const.CodeEnum.OK:
                continue
            n, _, code = await core.node.update_md(
                au=au,
                nid=nid,
                md=n["md"],
                refresh_on_same_md=True,
            )
            if code != const.CodeEnum.OK:
                continue
        if count % 20 == 0:
            doc, code = await utils.update_process(
                uid=uid, type_=type_, process=int(count / len(doc["obsidian"]) * 10 + 90))
            if code != const.CodeEnum.OK:
                await utils.set_running_false(
                    uid,
                    code,
                    msg="uploading process update failed",
                )
                logger.error(f"error: {code}, uid: {uid}")
                return
            if not doc["running"]:
                break

        count += 1

    t5 = time.time()
    logger.debug(f"obsidian upload, uid={uid}, update for old obsidian files time: {t5 - t4:.2f}")

    await utils.finish_task(uid=uid, obsidian=existed_path2nid)



import collections
import itertools
import threading
from sys import intern as itrn

import bson

from .engine_common import Engine
from ..common import MetaStorageObject


class MemoryEngine(Engine):
    def __init__(self, strict=False):
        self._strict = strict
        self._cache = collections.defaultdict(dict)
        self._metadata = {}
        self.lock = threading.RLock()

    @staticmethod
    def create(strict=False):
        return MemoryEngine(strict)

    def put_doc(self, collection, doc, no_overwrite=False):
        doc_id = str(doc['_id'])
        if no_overwrite and doc_id in self._cache[collection]:
            return False
        if self._strict:
            self._cache[itrn(collection)][itrn(doc_id)] = bson.encode(doc)
        else:
            self._cache[itrn(collection)][itrn(doc_id)] = doc
        return True

    def get_doc(self, collection, doc_id):
        obj = self._cache[collection].get(str(doc_id))
        if self._strict and obj:
            return bson.decode(obj)
        return obj

    def doc_exists(self, collection, doc_id):
        return str(doc_id) in self._cache[collection]

    def list_ids(self, collection, limit=None):
        keys = self._cache.get(collection, {}).keys()
        if limit is None:
            return list(keys)
        return list(itertools.islice(keys, limit))

    def delete_doc(self, collection, doc_id):
        self._cache[collection].pop(str(doc_id), None)
        return True

    def delete_dir(self, collection):
        with self.lock:
            self._cache.pop(collection, None)
            self._metadata.pop(collection, None)

        return True

    def put_metadata(self, collection, doc):
        self._metadata[collection] = doc.to_storage(as_bson=self._strict)
        return True

    def get_metadata(self, collection):
        try:
            obj = self._metadata[collection]
        except KeyError:
            return None
        return MetaStorageObject.from_storage(obj, from_bson=self._strict)

    def create_path(self, collection):
        pass

    def close(self):
        self._cache = collections.defaultdict(dict)
        self._metadata = {}



import copy
import datetime
from typing import List, Optional, Tuple, Dict, Any

from bson import ObjectId
from bson.tz_util import utc

from retk import config, const, utils, regex
from retk import plugins
from retk.core import user
from retk.logger import logger
from retk.models import tps, db_ops
from retk.models.client import client
from retk.models.search_engine.engine import SearchDoc
from . import backup, node_utils


@plugins.handler.on_node_added
async def post(  # noqa: C901
        au: tps.AuthedUser,
        md: str,
        type_: int = const.NodeTypeEnum.MARKDOWN.value,
        from_nid: str = "",
) -> Tuple[Optional[tps.Node], const.CodeEnum]:
    md = md.strip()
    if len(md) > const.settings.MD_MAX_LENGTH:
        return None, const.CodeEnum.NOTE_EXCEED_MAX_LENGTH

    new_size = len(md.encode("utf-8"))
    if await user.user_space_not_enough(au=au):
        return None, const.CodeEnum.USER_SPACE_NOT_ENOUGH

    title, body, snippet = utils.preprocess_md(md)

    nid = utils.short_uuid()

    from_nids = []
    if from_nid != "":
        from_nids.append(from_nid)
        res = await db_ops.node_add_to_set(from_nid, "toNodeIds", nid)
        if res.modified_count != 1:
            return None, const.CodeEnum.OPERATION_FAILED

    new_to_node_ids = []
    if md != "":
        new_to_node_ids, code = await node_utils.flush_to_node_ids(nid=nid, orig_to_nid=[], new_md=md)
        if code != const.CodeEnum.OK:
            return None, code
    _id = ObjectId()
    data = utils.get_node_dict(
        _id=_id,
        nid=nid,
        uid=au.u.id,
        md=md,
        title=title,
        snippet=snippet,
        type_=type_,
        disabled=False,
        in_trash=False,
        modified_at=_id.generation_time,
        in_trash_at=None,
        from_node_ids=from_nids,
        to_node_ids=new_to_node_ids,
        history=[],
    )

    res = await client.coll.nodes.insert_one(data)
    if not res.acknowledged:
        return None, const.CodeEnum.OPERATION_FAILED

    await user.update_used_space(uid=au.u.id, delta=new_size)

    code = await backup.storage_md(node=data, keep_hist=False)
    if code != const.CodeEnum.OK:
        return data, code

    if type_ == const.NodeTypeEnum.MARKDOWN.value:
        code = await client.search.add(au=au, doc=SearchDoc(nid=nid, title=title, body=body))
        if code != const.CodeEnum.OK:
            logger.error(f"add search index failed, code: {code}")
    return data, const.CodeEnum.OK


async def get(
        au: tps.AuthedUser,
        nid: str,
        with_disabled: bool = False,
        in_trash: bool = False,
) -> Tuple[Optional[tps.Node], const.CodeEnum]:
    docs, code = await get_batch(
        au=au,
        nids=[nid],
        with_disabled=with_disabled,
        in_trash=in_trash,
    )
    return docs[0] if len(docs) > 0 else None, code


async def get_batch(
        au: tps.AuthedUser,
        nids: List[str],
        with_disabled: bool = False,
        in_trash: bool = False,
) -> Tuple[List[tps.Node], const.CodeEnum]:
    for nid in nids:
        if regex.NID.match(nid) is None:
            logger.error(f"invalid nid: {nid}")
            return [], const.CodeEnum.NODE_NOT_EXIST
    c: Dict[str, Any] = {"uid": au.u.id, "inTrash": in_trash}
    if len(nids) > 1:
        c["id"] = {"$in": nids}
    elif len(nids) == 1:
        c["id"] = nids[0]
    if not with_disabled:
        c["disabled"] = False
    docs = await client.coll.nodes.find(c).to_list(length=None)
    if len(docs) != len(nids):
        logger.error(f"docs len != nids len: {nids}")
        return [], const.CodeEnum.NODE_NOT_EXIST

    await node_utils.set_linked_nodes(
        docs=docs,
        with_disabled=with_disabled,
    )
    return docs, const.CodeEnum.OK


@plugins.handler.on_node_updated
@plugins.handler.before_node_updated
async def update_md(  # noqa: C901
        au: tps.AuthedUser,
        nid: str,
        md: str,
        refresh_on_same_md: bool = False,
) -> Tuple[Optional[tps.Node], Optional[tps.Node], const.CodeEnum]:
    """

    Args:
        au:
        nid:
        md:
        refresh_on_same_md:

    Returns:
        Tuple[Optional[tps.Node], Optional[tps.Node], const.CodeEnum]: new node, old node, code
    """
    if regex.NID.match(nid) is None:
        return None, None, const.CodeEnum.NODE_NOT_EXIST
    md = md.strip()
    if len(md) > const.settings.MD_MAX_LENGTH:
        return None, None, const.CodeEnum.NOTE_EXCEED_MAX_LENGTH
    if await user.user_space_not_enough(au=au):
        return None, None, const.CodeEnum.USER_SPACE_NOT_ENOUGH

    title, body, snippet = utils.preprocess_md(md)

    n, code = await get(au=au, nid=nid)
    if code != const.CodeEnum.OK:
        return None, None, code
    old_n = copy.deepcopy(n)
    if n["md"] == md and not refresh_on_same_md:
        return n, old_n, const.CodeEnum.OK

    old_md_size = len(n["md"].encode("utf-8"))
    new_data = {
        "modifiedAt": datetime.datetime.now(tz=utc),
    }

    if n["title"] != title:
        # update it's title in fromNodes md's link
        from_nodes = await client.coll.nodes.find({"id": {"$in": n["fromNodeIds"]}}).to_list(length=None)
        for from_node in from_nodes:
            new_md = utils.change_link_title(md=from_node["md"], nid=nid, new_title=title)
            n, old_n, code = await update_md(au=au, nid=from_node["id"], md=new_md)
            if code != const.CodeEnum.OK:
                logger.error(f"update fromNode {from_node['id']} failed")
        new_data["title"] = title

    if n["md"] != md:
        new_data["md"] = md
        if n["snippet"] != snippet:
            new_data["snippet"] = snippet

    new_data["toNodeIds"], code = await node_utils.flush_to_node_ids(
        nid=n["id"], orig_to_nid=n["toNodeIds"], new_md=md)
    if code != const.CodeEnum.OK:
        return None, old_n, code

    if not config.is_local_db():
        doc = await client.coll.nodes.find_one_and_update(
            {"id": nid},
            {"$set": new_data},
            return_document=True,  # return updated doc
        )
    else:
        # local db not support find_one_and_update
        res = await client.coll.nodes.update_one(
            {"id": nid},
            {"$set": new_data}
        )
        if res.modified_count != 1:
            logger.error(f"update node {nid} failed")
            return None, old_n, const.CodeEnum.OPERATION_FAILED
        doc = await client.coll.nodes.find_one({"id": nid})

    if doc is None:
        return None, old_n, const.CodeEnum.NODE_NOT_EXIST
    await node_utils.set_linked_nodes(
        docs=[doc],
        with_disabled=False,
    )

    await user.update_used_space(uid=au.u.id, delta=len(md.encode("utf-8")) - old_md_size)

    if doc["type"] == const.NodeTypeEnum.MARKDOWN.value:
        code = await client.search.update(au=au, doc=SearchDoc(nid=nid, title=title, body=body))
        if code != const.CodeEnum.OK:
            logger.error(f"update search index failed, code: {code}")

    code = await backup.storage_md(node=doc, keep_hist=True)
    if code != const.CodeEnum.OK:
        return doc, old_n, code
    return doc, old_n, code


async def to_trash(au: tps.AuthedUser, nid: str) -> const.CodeEnum:
    return await batch_to_trash(au=au, nids=[nid])


async def batch_to_trash(au: tps.AuthedUser, nids: List[str]) -> const.CodeEnum:
    ns, code = await get_batch(au=au, nids=nids, with_disabled=True, in_trash=False)
    if code != const.CodeEnum.OK:
        return code

    recent_cursor_search_selected_nids = au.u.last_state.recent_cursor_search_selected_nids
    changed = False
    for n in ns:
        try:
            recent_cursor_search_selected_nids.remove(n["id"])
            changed = True
        except ValueError:
            pass
    if changed:
        res = await client.coll.users.update_one({"id": au.u.id}, {"$set": {
            "lastState.recentCursorSearchSelectedNIds": recent_cursor_search_selected_nids
        }})
        if res.modified_count != 1:
            logger.error(f"update user {au.u.id} failed")
            return const.CodeEnum.OPERATION_FAILED
    res = await client.coll.nodes.update_many({"id": {"$in": nids}}, {"$set": {
        "inTrash": True,
        "inTrashAt": datetime.datetime.now(tz=utc)
    }})
    if res.modified_count != len(nids):
        logger.error(f"update nodes {nids} failed")
        return const.CodeEnum.OPERATION_FAILED

    code = await client.search.batch_to_trash(au=au, nids=nids)
    if code != const.CodeEnum.OK:
        logger.error(f"update search index failed, code: {code}")
    return code


async def get_nodes_in_trash(
        au: tps.AuthedUser,
        page: int,
        limit: int
) -> Tuple[List[tps.Node], int]:
    condition = {
        "uid": au.u.id,
        "disabled": False,
        "inTrash": True,
    }
    docs = client.coll.nodes.find(condition).sort([("inTrashAt", -1), ("_id", -1)])
    total = await client.coll.nodes.count_documents(condition)
    if limit > 0:
        docs = docs.skip(page * limit).limit(limit)

    return await docs.to_list(length=None), total


async def restore_from_trash(au: tps.AuthedUser, nid: str) -> const.CodeEnum:
    return await restore_batch_from_trash(au=au, nids=[nid])


async def restore_batch_from_trash(au: tps.AuthedUser, nids: List[str]) -> const.CodeEnum:
    # restore nodes
    res = await client.coll.nodes.update_many({"id": {"$in": nids}}, {"$set": {
        "inTrash": False,
        "inTrashAt": None,
    }})
    if res.modified_count != len(nids):
        logger.error(f"restore nodes {nids} failed")
        return const.CodeEnum.OPERATION_FAILED

    code = await client.search.restore_batch_from_trash(au=au, nids=nids)
    if code != const.CodeEnum.OK:
        logger.error(f"restore search index failed, code: {code}")
    return code


async def delete(au: tps.AuthedUser, nid: str) -> const.CodeEnum:
    return await batch_delete(au=au, nids=[nid])


async def batch_delete(au: tps.AuthedUser, nids: List[str]) -> const.CodeEnum:
    ns, code = await get_batch(au=au, nids=nids, with_disabled=True, in_trash=True)
    if code != const.CodeEnum.OK:
        return code

    # remove fromNodes for linked nodes, not necessary
    # for n in ns:
    #     for linked_nid in n["fromNodeIds"]:
    #         db_ops.remove_to_node(from_nid=linked_nid, to_nid=n["id"])

    used_space_delta = 0
    # remove toNodes for linked nodes
    for n in ns:
        used_space_delta -= len(n["md"].encode("utf-8"))
        for linked_nid in n["toNodeIds"]:
            await db_ops.remove_from_node(from_nid=n["id"], to_nid=linked_nid)

    # delete user file
    await user.update_used_space(uid=au.u.id, delta=used_space_delta)

    # remove node
    res = await client.coll.nodes.delete_many({"id": {"$in": nids}, "uid": au.u.id})
    if res.deleted_count != len(nids):
        logger.error(f"delete nodes {nids} failed")
        return const.CodeEnum.OPERATION_FAILED

    backup.delete_node_md(uid=au.u.id, nids=nids)

    code = await client.search.delete_batch(au=au, nids=nids)
    if code != const.CodeEnum.OK:
        logger.error(f"delete search index failed, code: {code}")

    return code


async def disable(
        au: tps.AuthedUser,
        nid: str,
) -> const.CodeEnum:
    if regex.NID.match(nid) is None:
        return const.CodeEnum.NODE_NOT_EXIST
    res = await client.coll.nodes.update_one(
        {"id": nid},
        {"$set": {"disabled": True}}
    )
    if res.modified_count != 1:
        logger.error(f"disable node {nid} failed")
        return const.CodeEnum.OPERATION_FAILED
    code = await client.search.disable(au=au, nid=nid)
    if code != const.CodeEnum.OK:
        logger.error(f"disable search index failed, code: {code}")
    return const.CodeEnum.OK


async def core_nodes(
        au: tps.AuthedUser,
        page: int,
        limit: int,
) -> Tuple[List[tps.Node], int]:
    condition = {
        "uid": au.u.id,
        "disabled": False,
        "inTrash": False,
    }
    # the key of toNodeIds is a list, sort by the toNodeIds length, from large to small
    docs = db_ops.sort_nodes_by_to_nids(condition=condition, page=page, limit=limit)
    total = await client.coll.nodes.count_documents(condition)
    return await docs.to_list(length=None), total


async def new_user_add_default_nodes(language: str, uid: str) -> const.CodeEnum:
    lns = const.NEW_USER_DEFAULT_NODES[language]
    u = await client.coll.users.find_one({"id": uid})
    if u is None:
        return const.CodeEnum.USER_NOT_EXIST
    au = tps.AuthedUser(u=tps.convert_user_dict_to_authed_user(u), language=language, request_id="")
    n, code = await post(
        au=au,
        md=lns[0],
    )
    if code != const.CodeEnum.OK:
        return code
    _, code = await post(
        au=au,
        md=lns[1].format(n["id"]),
    )
    await client.search.refresh()
    return code


async def get_hist_editions(
        au: tps.AuthedUser,
        nid: str
) -> Tuple[List[str], const.CodeEnum]:
    """

    Args:
        au:
        nid:

    Returns:
        Tuple[List[str], const.CodeEnum]: history, code
    """
    n, code = await get(au=au, nid=nid)
    if code != const.CodeEnum.OK:
        return [], code
    return n.get("history", []), const.CodeEnum.OK


async def get_hist_edition_md(au: tps.AuthedUser, nid: str, version: str) -> Tuple[str, const.CodeEnum]:
    n, code = await get(au=au, nid=nid)
    if code != const.CodeEnum.OK:
        return "", code
    if version not in n["history"]:
        return "", const.CodeEnum.NODE_NOT_EXIST
    return backup.get_md(uid=au.u.id, nid=nid, version=version)



from typing import List

from pydantic import Field, BaseModel

from retk.const import settings


class AtNodeRequest(BaseModel):
    nid: str = Field(max_length=settings.NID_MAX_LENGTH)
    toNid: str = Field(max_length=settings.NID_MAX_LENGTH)


class GetRecentSearchResponse(BaseModel):
    requestId: str
    queries: List[str]



from typing import List, Set, Tuple

from retk import const, regex
from retk.models import tps, db_ops
from retk.models.client import client


def get_linked_nodes(new_md) -> Tuple[set, const.CodeEnum]:
    # last first
    cache_current_to_nid: Set[str] = set()

    for match in list(regex.MD_AT_LINK.finditer(new_md))[::-1]:
        l0, l1 = match.span(1)
        link = new_md[l0:l1]
        if link.startswith("/n/"):
            # check existed nodes
            to_nid = link[3:]
            cache_current_to_nid.add(to_nid)
    return cache_current_to_nid, const.CodeEnum.OK


async def flush_to_node_ids(
        nid: str,
        orig_to_nid: List[str],
        new_md: str
) -> Tuple[List[str], const.CodeEnum]:
    new_to_nid, code = get_linked_nodes(new_md=new_md)
    if code != const.CodeEnum.OK:
        return [], code

    # remove fromNodes for linked nodes
    orig_to_nid = set(orig_to_nid)
    for to_nid in orig_to_nid.difference(new_to_nid):
        await db_ops.remove_from_node(from_nid=nid, to_nid=to_nid)

    # add fromNodes for linked nodes
    for to_nid in new_to_nid.difference(orig_to_nid):
        await db_ops.node_add_to_set(id_=to_nid, key="fromNodeIds", value=nid)

    return list(new_to_nid), const.CodeEnum.OK


async def set_linked_nodes(
        docs: List[tps.Node],
        with_disabled: bool = False,
):
    for doc in docs:
        doc["fromNodes"] = await client.coll.nodes.find({
            "id": {"$in": doc["fromNodeIds"]},
            "disabled": with_disabled,
        }).to_list(length=None)
        doc["toNodes"] = await client.coll.nodes.find({
            "id": {"$in": doc["toNodeIds"]},
            "disabled": with_disabled,
        }).to_list(length=None)



import datetime
import logging
from typing import List, Tuple, Sequence, Literal

import jieba
from bson.tz_util import utc
from jieba.analyse import ChineseAnalyzer
from whoosh.fields import TEXT, ID, Schema, DATETIME, BOOLEAN
from whoosh.highlight import Highlighter, HtmlFormatter
from whoosh.index import create_in, open_dir, FileIndex
from whoosh.qparser import QueryParser, syntax
from whoosh.query import Term, And, Or, Every

from retk import config, const
from retk.logger import logger
from retk.models.search_engine.engine import (
    BaseEngine, SearchDoc, SearchResult, RestoreSearchDoc, STOPWORDS,
)
from retk.models.tps import AuthedUser

jieba.setLogLevel(logging.ERROR)


class LocalSearcher(BaseEngine):
    ix: FileIndex
    indexing_schema: Schema
    search_stop_schema: Schema

    async def _trash_disable_ops_batch(
            self,
            au: AuthedUser,
            nids: List[str],
            disable: bool = None,
            in_trash: bool = None
    ) -> const.CodeEnum:
        writer = self.ix.writer()
        for nid in nids:
            with self.ix.searcher() as searcher:
                resp = searcher.search(And([Term("uid", au.u.id), Term("nid", nid)]))
                if len(resp) != 1:
                    logger.error(f"nid {nid} not found or more than one found")
                    return const.CodeEnum.NODE_NOT_EXIST
                res = resp[0]
                new = {
                    "createdAt": res["createdAt"],
                    "nid": nid,
                    "modifiedAt": res["modifiedAt"],
                    "title": res["title"],
                    "body": res["body"],
                    "disabled": disable if disable is not None else res["disabled"],
                    "inTrash": in_trash if in_trash is not None else res["inTrash"],
                }

            writer.update_document(uid=au.u.id, **new)
        writer.commit()
        return const.CodeEnum.OK

    @property
    def index_path(self):
        return config.get_settings().RETHINK_LOCAL_STORAGE_PATH / const.settings.DOT_DATA / "search"

    async def init(self):
        def create_schema(analyzer):
            return Schema(
                uid=ID(stored=True),
                nid=ID(stored=True, unique=True),
                title=TEXT(analyzer=analyzer, stored=True, sortable=True),
                body=TEXT(analyzer=analyzer, stored=True),
                disabled=BOOLEAN(stored=True),
                inTrash=BOOLEAN(stored=True),
                modifiedAt=DATETIME(stored=True, sortable=True),
                createdAt=DATETIME(stored=True, sortable=True),
            )

        self.indexing_schema = create_schema(analyzer=ChineseAnalyzer())
        self.search_stop_schema = create_schema(analyzer=ChineseAnalyzer(stoplist=STOPWORDS))

        if not self.index_path.exists():
            self.index_path.mkdir(parents=True)
            self.ix = create_in(self.index_path, self.indexing_schema)
        else:
            self.ix = open_dir(self.index_path)

    async def close(self):
        self.ix.close()

    async def drop(self):
        try:
            w = self.ix.writer()
            w.delete_by_query(Every("nid"))
            w.commit()
            self.ix.close()
        except (AttributeError, FileNotFoundError):
            pass
        # # remove all index
        # if self.index_path.exists():
        #     rmtree(self.index_path)

    async def add(self, au: AuthedUser, doc: SearchDoc) -> const.CodeEnum:
        return await self.add_batch(au=au, docs=[doc])

    async def update(self, au: AuthedUser, doc: SearchDoc) -> const.CodeEnum:
        return await self.update_batch(au=au, docs=[doc])

    async def to_trash(self, au: AuthedUser, nid: str) -> const.CodeEnum:
        return await self.batch_to_trash(au=au, nids=[nid])

    async def restore_from_trash(self, au: AuthedUser, nid: str) -> const.CodeEnum:
        return await self.restore_batch_from_trash(au=au, nids=[nid])

    async def disable(self, au: AuthedUser, nid: str) -> const.CodeEnum:
        return await self._trash_disable_ops_batch(au=au, nids=[nid], disable=True)

    async def enable(self, au: AuthedUser, nid: str) -> const.CodeEnum:
        return await self._trash_disable_ops_batch(au=au, nids=[nid], disable=False)

    async def delete(self, au: AuthedUser, nid: str) -> const.CodeEnum:
        return await self.delete_batch(au=au, nids=[nid])

    async def add_batch(self, au: AuthedUser, docs: List[SearchDoc]) -> const.CodeEnum:
        writer = self.ix.writer()
        now = datetime.datetime.now(tz=utc)
        for doc in docs:
            d = doc.__dict__
            now_ = datetime.datetime.now(tz=utc)
            delta = datetime.timedelta(microseconds=100)
            if now + delta >= now_:
                now_ = now + delta
            d["createdAt"] = now_
            d["modifiedAt"] = d["createdAt"]
            d["disabled"] = False
            d["inTrash"] = False
            writer.add_document(uid=au.u.id, **d)
            now = now_

        writer.commit()
        return const.CodeEnum.OK

    async def batch_to_trash(self, au: AuthedUser, nids: List[str]) -> const.CodeEnum:
        return await self._trash_disable_ops_batch(au=au, nids=nids, in_trash=True)

    async def restore_batch_from_trash(self, au: AuthedUser, nids: List[str]) -> const.CodeEnum:
        return await self._trash_disable_ops_batch(au=au, nids=nids, in_trash=False)

    async def delete_batch(self, au: AuthedUser, nids: List[str]) -> const.CodeEnum:
        writer = self.ix.writer()
        for nid in nids:
            q = And([Term("uid", au.u.id), Term("nid", nid), Term("inTrash", True)])
            count = writer.delete_by_query(q=q)
            if count != 1:
                logger.error(f"nid {nid} not found or more than one found")
                return const.CodeEnum.NODE_NOT_EXIST
        writer.commit()
        return const.CodeEnum.OK

    async def force_delete_all(self, uid: str) -> const.CodeEnum:
        writer = self.ix.writer()
        q = Term("uid", uid)
        writer.delete_by_query(q=q)
        writer.commit()
        return const.CodeEnum.OK

    async def update_batch(self, au: AuthedUser, docs: List[SearchDoc]) -> const.CodeEnum:
        writer = self.ix.writer()
        for doc in docs:
            with self.ix.searcher() as searcher:
                resp = searcher.search(And([Term("uid", au.u.id), Term("nid", doc.nid)]))
                if len(resp) != 1:
                    logger.error(f"nid {doc.nid} not found or more than one found")
                    return const.CodeEnum.NODE_NOT_EXIST
                res = resp[0]
                new = {
                    "createdAt": res["createdAt"],
                    "nid": doc.nid,
                    "modifiedAt": datetime.datetime.now(tz=utc),
                    "title": doc.title if doc.title != "" else res["title"],
                    "body": doc.body if doc.body != "" else res["body"],
                    "disabled": res["disabled"],
                    "inTrash": res["inTrash"],
                }
            writer.update_document(uid=au.u.id, **new)
        writer.commit()
        return const.CodeEnum.OK

    async def _search(
            self,
            au: AuthedUser,
            query: str = "",
            sort_key: Literal[
                "createdAt", "modifiedAt", "title", "similarity"
            ] = None,
            reverse: bool = False,
            page: int = 0,
            page_size: int = 10,
            exclude_nids: Sequence[str] = None,
            with_stop_analyzer: bool = False,
    ) -> Tuple[List[SearchResult], int]:
        if sort_key in ["", "similarity"]:
            sort_key = None
        with self.ix.searcher() as searcher:
            cs = [Term("uid", au.u.id), Term("disabled", False), Term("inTrash", False)]
            if query != "":
                schema = self.search_stop_schema if with_stop_analyzer else self.indexing_schema
                q_body = QueryParser("body", schema, group=syntax.OrGroup).parse(query.lower())
                q_title = QueryParser("title", schema, group=syntax.OrGroup).parse(query.lower())
                cs.append(Or([q_body, q_title]))
            query_terms = And(cs)
            if exclude_nids is not None and len(exclude_nids) > 0:
                mask = Term("nid", exclude_nids)
            else:
                mask = None
            hits = searcher.search_page(
                query=query_terms,
                pagenum=page + 1,
                pagelen=page_size,
                sortedby=sort_key,
                reverse=reverse,
                mask=mask,
            )
            if page * page_size > hits.total:
                return [], hits.total
            hl = Highlighter(formatter=HtmlFormatter(
                tagname=self.hl_tag_name,
                classname=self.hl_class_name,
                termclass=self.hl_term_prefix
            ))

            return [
                SearchResult(
                    nid=hit["nid"],
                    score=hit.score if sort_key != "title" else 0.,
                    titleHighlight=self.get_hl(hl, hit, "title", return_list=False, default=hit["title"]),
                    bodyHighlights=self.get_hl(hl, hit, "body", return_list=True, default=hit["body"][:60] + "..."),
                ) for hit in hits
            ], hits.total

    async def search(
            self,
            au: AuthedUser,
            query: str = "",
            sort_key: Literal[
                "createdAt", "modifiedAt", "title", "similarity"
            ] = None,
            reverse: bool = False,
            page: int = 0,
            limit: int = 10,
            exclude_nids: Sequence[str] = None,
    ) -> Tuple[List[SearchResult], int]:
        return await self._search(
            au=au,
            query=query,
            sort_key=sort_key,
            reverse=reverse,
            page=page,
            page_size=limit,
            exclude_nids=exclude_nids,
            with_stop_analyzer=False,
        )

    async def recommend(
            self,
            au: AuthedUser,
            content: str,
            max_return: int = 10,
            exclude_nids: Sequence[str] = None,
    ) -> List[SearchResult]:
        threshold = 5.
        res, _ = await self._search(
            au=au,
            query=content,
            sort_key="similarity",
            reverse=True,
            page=0,
            page_size=max_return,
            exclude_nids=exclude_nids,
            with_stop_analyzer=True,
        )
        return [r for r in res if r.score >= threshold]

    async def refresh(self):
        pass

    async def count_all(self) -> int:
        with self.ix.searcher() as searcher:
            resp = searcher.search(Every("nid"))
            return len(resp)

    async def batch_restore_docs(self, au: AuthedUser, docs: List[RestoreSearchDoc]) -> const.CodeEnum:
        writer = self.ix.writer()
        for doc in docs:
            writer.add_document(uid=au.u.id, **doc.__dict__)
        writer.commit()
        return const.CodeEnum.OK

    @staticmethod
    def get_hl(hl: Highlighter, hit: dict, key: str, return_list: bool, default: str = ""):
        hl_str = hl.highlight_hit(hit, key)
        if return_list:
            if hl_str == "":
                return [default]
            return [hl_str]
        if hl_str == "":
            return default
        return hl_str



from retk import const, config
from retk.controllers import schemas
from retk.controllers.utils import maybe_raise_json_exception, json_exception
from retk.core import node
from retk.models.tps import AuthedUser
from retk.plugins.base import get_plugins, event_plugin_map


async def get_all_plugins(
        au: AuthedUser,
) -> schemas.plugin.PluginsResponse:
    plugins = [
        schemas.plugin.PluginsResponse.Plugin(
            id=p.id,
            name=p.name,
            version=p.version,
            description=p.description,
            author=p.author,
            iconSrc=p.icon_src,
        )
        for p in get_plugins().values()
    ]
    return schemas.plugin.PluginsResponse(
        requestId=au.request_id,
        plugins=plugins
    )


async def get_plugins_with_render_editor_side(
        au: AuthedUser,
) -> schemas.plugin.PluginsResponse:
    plugins = []
    for p in event_plugin_map["render_editor_side"]:
        plugins.append(
            schemas.plugin.PluginsResponse.Plugin(
                id=p.id,
                name=p.name,
                version=p.version,
                description=p.description,
                author=p.author,
                iconSrc=p.icon_src,
            )
        )
    return schemas.plugin.PluginsResponse(
        requestId=au.request_id,
        plugins=plugins
    )


async def __render(
        au: AuthedUser,
        pid: str,
        nid: str = "",
):
    plugins = get_plugins()
    if pid not in plugins:
        return maybe_raise_json_exception(au=au, code=const.CodeEnum.PLUGIN_NOT_FOUND)

    try:
        plugin = plugins[pid]
    except KeyError:
        return maybe_raise_json_exception(au=au, code=const.CodeEnum.PLUGIN_NOT_FOUND)

    try:
        if nid != "":
            n, code = await node.get(au=au, nid=nid)
            maybe_raise_json_exception(au=au, code=code)

            html = plugin.render_editor_side(
                uid=au.u.id, nid=nid, md=n["md"], language=au.language
            )
        else:
            html = plugin.render_plugin_home(language=au.language)

    except NotImplementedError:
        html = plugin.description
    return schemas.plugin.RenderPluginResponse(
        requestId=au.request_id,
        html=html,
    )


async def render_plugin_home(
        au: AuthedUser,
        pid: str,
) -> schemas.plugin.RenderPluginResponse:
    return await __render(au=au, pid=pid)


async def render_editor_side(
        au: AuthedUser,
        pid: str,
        nid: str,
) -> schemas.plugin.RenderPluginResponse:
    return await __render(au=au, pid=pid, nid=nid)


async def plugin_call(
        req: schemas.plugin.PluginCallRequest,
) -> schemas.plugin.PluginCallResponse:
    if not config.is_local_db():
        raise json_exception(
            request_id=req.requestId,
            code=const.CodeEnum.NOT_PERMITTED,
            language=const.LanguageEnum.EN.value,
            log_msg="plugin call is not allowed in production",
        )
    plugins = get_plugins()
    try:
        plugin = plugins[req.pluginId]
    except KeyError:
        return schemas.plugin.PluginCallResponse(
            success=False,
            message="plugin not found",
            requestId=req.requestId,
            pluginId=req.pluginId,
            method=req.method,
            data=None,
        )

    res = plugin.handle_api_call(req.method, req.data)
    return schemas.plugin.PluginCallResponse(
        success=res.success,
        message=res.message,
        requestId=req.requestId,
        pluginId=req.pluginId,
        method=req.method,
        data=res.data,
    )



"""Github SSO Oauth Helper class"""

from .base import DiscoveryDocument, OpenID, SSOBase


class GithubSSO(SSOBase):
    """Class providing login via Github SSO"""

    provider = "github"
    scope = ["user:email"]
    additional_headers = {"accept": "application/json"}

    async def get_discovery_document(self) -> DiscoveryDocument:
        return {
            "authorization_endpoint": "https://github.com/login/oauth/authorize",
            "token_endpoint": "https://github.com/login/oauth/access_token",
            "userinfo_endpoint": "https://api.github.com/user",
        }

    @classmethod
    async def openid_from_response(cls, response: dict) -> OpenID:
        return OpenID(
            email=response["email"] if response["email"] else "",
            provider=cls.provider,
            id=str(response["id"]),
            display_name=response["login"],
            picture=response["avatar_url"],
        )



import datetime
import io
import os
import shutil
import time
import unittest
from pathlib import Path
from typing import Dict
from unittest.mock import patch
from zipfile import ZipFile

from PIL import Image
from bson.tz_util import utc
from fastapi.testclient import TestClient
from httpx import Response

from retk import const, config, PluginAPICallReturn
from retk.application import app
from retk.core import account, scheduler
from retk.models.client import client
from retk.models.tps import convert_user_dict_to_authed_user
from retk.plugins.register import register_official_plugins, unregister_official_plugins
from retk.utils import jwt_decode
from . import utils


class PublicApiTest(unittest.IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        utils.set_env(".env.test.local")

    async def asyncSetUp(self) -> None:
        await client.init()
        self.client = TestClient(app)

    async def asyncTearDown(self) -> None:
        await client.drop()

    @classmethod
    def tearDownClass(cls) -> None:
        utils.drop_env(".env.test.local")

    def test_home(self):
        resp = self.client.get("/")
        self.assertEqual(200, resp.status_code)
        self.assertEqual("text/html; charset=utf-8", resp.headers["content-type"])
        self.assertEqual(Response, type(resp))

    def test_register(self):
        resp = self.client.put(
            "/api/account/auto-login",
            headers={"RequestId": "xxx"},
        )
        self.assertEqual(200, resp.status_code)
        rj = resp.json()
        self.assertEqual("xxx", rj["requestId"])
        self.assertIsNone(rj["user"])

        token, _ = account.app_captcha.generate()
        data = jwt_decode(token)
        resp = self.client.post(
            "/api/account",
            json={
                "email": "a@c.com",
                "password": "a",
                "verificationToken": token,
                "verification": data["code"],
                "language": const.LanguageEnum.EN.value,
                "requestId": "xxx"
            },
            headers={"RequestId": "xxx"}
        )
        self.assertEqual(403, resp.status_code)
        rj = resp.json()
        self.assertEqual(const.CodeEnum.ONE_USER_MODE.value, rj["detail"]["code"])
        self.assertEqual("xxx", rj["detail"]["requestId"])

    @patch(
        "retk.core.account.email.EmailServer._send"
    )
    def test_email_verification(self, mock_send):
        mock_send.return_value = const.CodeEnum.OK
        token, _ = account.app_captcha.generate()
        data = jwt_decode(token)

        resp = self.client.put(
            "/api/account/email/send-code",
            json={
                "email": "a@c.com",
                "userExistOk": True,
                "captchaToken": token,
                "captchaCode": data["code"],
                "language": "zh",
            },
            headers={"RequestId": "xxx"}
        )
        self.assertEqual(200, resp.status_code)
        rj = resp.json()
        self.assertNotEqual("", rj["token"])
        self.assertEqual("xxx", rj["requestId"])


class TokenApiTest(unittest.IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        utils.set_env(".env.test.local")

    async def asyncSetUp(self) -> None:
        scheduler.start()
        await client.init()
        self.client = TestClient(app)
        resp = self.client.put(
            "/api/account/login",
            json={
                "email": const.DEFAULT_USER["email"],
                "password": "",
            })
        self.assertEqual(200, resp.status_code, msg=resp.json())
        self.access_token = resp.cookies.get(const.settings.COOKIE_ACCESS_TOKEN)
        self.refresh_token_id = resp.cookies.get(const.settings.COOKIE_REFRESH_TOKEN_ID)
        self.refresh_token = resp.cookies.get(const.settings.COOKIE_REFRESH_TOKEN)

        self.assertEqual(842, len(self.access_token))
        self.assertTrue(self.access_token.startswith("\"Bearer "), msg=self.access_token)
        self.assertTrue(self.refresh_token.startswith("\"Bearer "), msg=self.access_token)
        self.default_headers = {
            "RequestId": "xxx",
        }

        resp = self.client.get(
            "/api/users",
            headers=self.default_headers,
        )
        rj = resp.json()
        self.assertEqual(200, resp.status_code, msg=rj)
        self.assertNotIn("detail", rj)

    async def asyncTearDown(self) -> None:
        config.get_settings().ONE_USER = True
        self.client.cookies.clear()
        scheduler.stop()
        await client.drop()
        self.client.close()
        shutil.rmtree(Path(__file__).parent / "tmp" / const.settings.DOT_DATA / "files", ignore_errors=True)
        shutil.rmtree(Path(__file__).parent / "tmp" / const.settings.DOT_DATA / "md", ignore_errors=True)

    @classmethod
    def tearDownClass(cls) -> None:
        utils.drop_env(".env.test.local")

    def error_check(self, resp: Response, status_code: int, code: const.CodeEnum, rid="xxx", language="en"):
        self.assertEqual(status_code, resp.status_code)
        rj = resp.json()
        self.assertIn("detail", rj, msg=rj)
        detail = rj["detail"]
        self.assertEqual(code.value, detail["code"], msg=detail)
        self.assertEqual(rid, detail["requestId"], msg=detail)
        self.assertEqual(const.get_msg_by_code(code, language=language), detail["msg"], msg=detail)

    def check_ok_response(self, resp: Response, status_code: int = 200, rid="xxx") -> dict:
        self.assertEqual(status_code, resp.status_code, msg=resp.json())
        rj = resp.json()
        self.assertEqual(rid, rj["requestId"])
        return rj

    def set_access_token(self, token: str):
        try:
            self.client.cookies.delete(const.settings.COOKIE_ACCESS_TOKEN)
        except KeyError:
            pass
        self.client.cookies[const.settings.COOKIE_ACCESS_TOKEN] = token

    async def create_new_temp_user(self, email):
        config.get_settings().ONE_USER = False
        config.get_settings().DB_SALT = "test"
        token, code = account.app_captcha.generate()
        data = jwt_decode(token)
        code = data["code"].replace(config.get_settings().CAPTCHA_SALT, "")
        lang = "zh"

        resp = self.client.post(
            "/api/account",
            json={
                "email": email,
                "password": "abc111",
                "verificationToken": token,
                "verification": code,
                "language": lang,
            },
            headers={"RequestId": "xxx"}
        )
        self.check_ok_response(resp, 201)
        return resp

    async def set_default_manager(self):
        u = await client.coll.users.find_one({"email": const.DEFAULT_USER["email"]})
        admin_uid = u["id"]
        doc = await client.coll.users.update_one(
            {"id": admin_uid},
            {"$set": {"type": const.USER_TYPE.ADMIN.id}}
        )
        self.assertEqual(1, doc.modified_count)
        return admin_uid

    async def clear_default_manager(self, admin_uid):
        doc = await client.coll.users.update_one(
            {"id": admin_uid},
            {"$set": {"type": const.USER_TYPE.NORMAL.id}}
        )
        self.assertEqual(1, doc.modified_count)

    async def test_auto_login(self):
        resp = self.client.put(
            "/api/account/auto-login",
            headers={"RequestId": "xxx"},
        )
        self.assertEqual(200, resp.status_code)
        rj = resp.json()
        self.assertEqual("xxx", rj["requestId"])
        self.assertIsNotNone(rj["user"])

        self.client.cookies.delete(const.settings.COOKIE_ACCESS_TOKEN)
        resp = self.client.put(
            "/api/account/auto-login",
            headers={"RequestId": "xxx"},
        )
        self.assertEqual(200, resp.status_code)
        self.assertIsNone(resp.json()["user"])

    async def test_access_refresh_token(self):
        self.client.put(
            "/api/account/login",
            json={
                "email": const.DEFAULT_USER["email"],
                "password": "",
            })

        resp = self.client.get(
            "/api/users",
            headers=self.default_headers,
        )
        self.assertEqual(200, resp.status_code)

        self.set_access_token(resp.cookies.get(const.settings.COOKIE_REFRESH_TOKEN))
        resp = self.client.get(
            "/api/users",
            headers=self.default_headers,
        )
        self.error_check(resp, 200, const.CodeEnum.EXPIRED_OR_NO_ACCESS_TOKEN)

    async def test_access_token_expire(self):
        aed = config.get_settings().ACCESS_TOKEN_EXPIRE_DELTA
        red = config.get_settings().REFRESH_TOKEN_EXPIRE_DELTA

        config.get_settings().ACCESS_TOKEN_EXPIRE_DELTA = datetime.timedelta(microseconds=1)
        config.get_settings().REFRESH_TOKEN_EXPIRE_DELTA = datetime.timedelta(microseconds=2)
        resp = self.client.put(
            "/api/account/login",
            json={
                "email": const.DEFAULT_USER["email"],
                "password": "",
                "language": "zh",
            })
        self.assertEqual(200, resp.status_code)
        time.sleep(0.001)

        resp = self.client.get(
            "/api/users",
            headers={
                "RequestId": "xxx"
            },
        )
        self.assertEqual(200, resp.status_code)
        rj = resp.json()
        detail = rj["detail"]
        self.assertEqual(const.CodeEnum.EXPIRED_OR_NO_ACCESS_TOKEN.value, detail["code"], msg=detail)
        self.assertEqual("xxx", detail["requestId"], msg=detail)

        resp = self.client.get(
            "/api/account/access-token",
            headers=self.default_headers,
        )
        self.error_check(resp, 401, const.CodeEnum.EXPIRED_AUTH)

        config.get_settings().REFRESH_TOKEN_EXPIRE_DELTA = red

        self.client.put(
            "/api/account/login",
            json={
                "email": const.DEFAULT_USER["email"],
                "password": "",
            })

        time.sleep(0.001)

        resp = self.client.get(
            "/api/users",
            headers=self.default_headers,
        )
        self.assertEqual(200, resp.status_code)
        rj = resp.json()
        self.assertEqual(const.CodeEnum.EXPIRED_OR_NO_ACCESS_TOKEN.value, rj["detail"]["code"], msg=rj)

        config.get_settings().ACCESS_TOKEN_EXPIRE_DELTA = aed
        old_access_token = resp.cookies.get(const.settings.COOKIE_ACCESS_TOKEN)
        resp = self.client.get(
            "/api/account/access-token",
            headers=self.default_headers,
        )
        self.check_ok_response(resp, 200)
        at = resp.cookies.get(const.settings.COOKIE_ACCESS_TOKEN)
        self.assertNotEqual(old_access_token, at)
        self.assertTrue(at.startswith("\"Bearer "))

        resp = self.client.get(
            "/api/users",
            headers=self.default_headers,
        )
        self.check_ok_response(resp, 200)
        rj = resp.json()
        self.assertEqual("rethink", rj["user"]["nickname"])

    async def test_add_user_update_password(self):
        config.get_settings().ONE_USER = False
        config.get_settings().DB_SALT = "test"
        token, code = account.app_captcha.generate()
        data = jwt_decode(token)
        code = data["code"].replace(config.get_settings().CAPTCHA_SALT, "")

        email = "a@b.c"
        del self.client.cookies[const.settings.COOKIE_ACCESS_TOKEN]
        resp = self.client.get(
            "/api/users",
            headers=self.default_headers,
        )
        self.error_check(resp, 200, const.CodeEnum.EXPIRED_OR_NO_ACCESS_TOKEN)

        lang = "zh"
        resp = self.client.post(
            "/api/account",
            json={
                "email": email,
                "password": "abc111",
                "verificationToken": token,
                "verification": code,
                "language": lang,
            },
            headers=self.default_headers
        )
        _ = self.check_ok_response(resp, 201)
        u_token = resp.cookies.get(const.settings.COOKIE_ACCESS_TOKEN)
        self.assertNotEqual("", u_token)

        resp = self.client.get(
            "/api/users",
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        self.assertEqual("a**@b.c", rj["user"]["email"])
        self.assertEqual("zh", rj["user"]["settings"]["language"])

        resp = self.client.put(
            "/api/users/password",
            json={
                "oldPassword": "xxx111",
                "newPassword": "abc222",
            },
            headers=self.default_headers,
        )
        self.error_check(resp, 400, const.CodeEnum.OLD_PASSWORD_ERROR, language=lang)

        resp = self.client.put(
            "/api/users/password", json={
                "email": email,
                "oldPassword": "abc111",
                "newPassword": "abc222",
            },
            headers=self.default_headers,
        )
        _ = self.check_ok_response(resp, 200)
        u = await client.coll.users.find_one({"email": email})
        au = convert_user_dict_to_authed_user(u)
        self.assertTrue(await account.manager.is_right_password(
            email=au.email,
            hashed=au.hashed,
            password="abc222"
        ))

        uid = (await client.coll.users.find_one({"email": email}))["id"]
        await client.coll.users.delete_one({"id": uid})
        await client.coll.nodes.delete_many({"uid": uid})
        config.get_settings().ONE_USER = True

    def test_get_user(self):
        resp = self.client.get(
            "/api/users",
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        self.assertEqual("rethink", rj["user"]["nickname"])
        self.assertGreater(
            datetime.datetime.strptime(
                rj["user"]["createdAt"], "%Y-%m-%dT%H:%M:%SZ"
            ).timestamp(),
            0., msg=rj["user"]["createdAt"])

    def test_update_user(self):
        resp = self.client.patch(
            "/api/users",
            json={
                "nickname": "new nickname",
                "avatar": "http://new.avatar/aa.png",
                "lastState": {
                    "nodeDisplayMethod": const.NodeDisplayMethodEnum.LIST.value,
                }
            },
            headers=self.default_headers,
        )
        _ = self.check_ok_response(resp, 200)

        resp = self.client.get(
            "/api/users",
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        self.assertEqual("new nickname", rj["user"]["nickname"])
        self.assertEqual("http://new.avatar/aa.png", rj["user"]["avatar"])
        self.assertEqual(const.NodeDisplayMethodEnum.LIST.value, rj["user"]["lastState"]["nodeDisplayMethod"])

        resp = self.client.patch(
            "/api/users",
            json={
                "settings": {
                    "language": const.LanguageEnum.ZH.value,
                    "theme": "dark",
                    "editorMode": "ir",
                    "editorFontSize": 20,
                    "editorCodeTheme": "dracula",
                }
            },
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        self.assertEqual("zh", rj["user"]["settings"]["language"])
        self.assertEqual("ir", rj["user"]["settings"]["editorMode"])
        self.assertEqual("dark", rj["user"]["settings"]["theme"])
        self.assertEqual(20, rj["user"]["settings"]["editorFontSize"])
        self.assertEqual("dracula", rj["user"]["settings"]["editorCodeTheme"])

    def test_recent_search(self):
        resp = self.client.get(
            "/api/nodes",
            params={
                "q": "aaa",
                "p": 0,
                "limit": 5
            },
            headers=self.default_headers,
        )
        self.check_ok_response(resp, 200)

        resp = self.client.get(
            "/api/recent/searched",
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        self.assertEqual(["aaa"], rj["queries"])

    def test_node(self):
        resp = self.client.get(
            "/api/nodes",
            params={
                "q": "",
                "sort": "createdAt",
                "ord": "desc",
                "p": 0,
                "limit": 5
            },
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        self.assertGreater(len(rj["data"]["nodes"]), 0)

        self.client.get(
            "/api/nodes",
            params={
                "q": "",
                "sort": "createdAt",
                "ord": "desc",
                "p": 0,
                "limit": 5
            },
            headers=self.default_headers,
        )
        self.check_ok_response(resp, 200)

        resp = self.client.post(
            "/api/nodes",
            json={
                "md": "node1\ntext",
                "type": const.NodeTypeEnum.MARKDOWN.value,
            },
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 201)
        node = rj["node"]

        resp = self.client.get(
            f'/api/nodes/{node["id"]}',
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        n = rj["node"]
        self.assertEqual("node1", n["title"], msg=rj)
        self.assertEqual("node1\ntext", n["md"])
        self.assertEqual(const.NodeTypeEnum.MARKDOWN.value, n["type"])

        resp = self.client.put(
            f'/api/nodes/{node["id"]}/md',
            json={
                "md": "node2\ntext",
            },
            headers=self.default_headers,
        )
        self.check_ok_response(resp, 200)

        resp = self.client.get(
            f'/api/nodes/{node["id"]}',
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        n = rj["node"]
        self.assertEqual("xxx", rj["requestId"])
        self.assertEqual("node2", n["title"])
        self.assertEqual("node2\ntext", n["md"])
        self.assertEqual(const.NodeTypeEnum.MARKDOWN.value, n["type"])

        resp = self.client.put(
            f'/api/trash/{node["id"]}',
            headers=self.default_headers,
        )
        self.check_ok_response(resp, 200)

        resp = self.client.get(
            "/api/trash",
            params={"p": 0, "limit": 10},
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        self.assertEqual(1, len(rj["data"]["nodes"]))

        resp = self.client.put(
            f'/api/trash/{node["id"]}/restore',
            headers=self.default_headers,
        )
        self.check_ok_response(resp, 200)

        resp = self.client.get(
            f'/api/nodes/{node["id"]}/at',
            params={
                "q": "How",
                "p": 0,
                "limit": 10,
            },
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        self.assertEqual(2, len(rj["data"]["nodes"]))
        self.assertEqual("Welcome to Rethink", rj["data"]["nodes"][0]["title"])

        resp = self.client.get(
            f'/api/nodes/{node["id"]}/recommend',
            params={
                "content": "I do need a Knowledge Management System. This is a good one to try.",
            },
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        self.assertEqual(1, len(rj["data"]["nodes"]))
        self.assertEqual(1, rj["data"]["total"])
        self.assertEqual("Welcome to Rethink", rj["data"]["nodes"][0]["title"])

        self.client.put(
            f'/api/trash/{node["id"]}',
            headers=self.default_headers
        )
        self.check_ok_response(resp, 200)
        resp = self.client.delete(
            f"/api/trash/{node['id']}",
            headers=self.default_headers,
        )
        self.assertEqual(200, resp.status_code)

        resp = self.client.delete(
            "/api/trash/ssa",
            headers=self.default_headers,
        )
        self.error_check(resp, 404, const.CodeEnum.NODE_NOT_EXIST)

        resp = self.client.get(
            f'/api/nodes/{node["id"]}',
            headers=self.default_headers,
        )
        self.error_check(resp, 404, const.CodeEnum.NODE_NOT_EXIST)

        resp = self.client.get(
            "/api/nodes/core",
            params={
                "p": 0,
                "limit": 10,
            },
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        self.assertEqual(2, rj["data"]["total"])
        self.assertEqual(2, len(rj["data"]["nodes"]))

    def test_captcha(self):
        resp = self.client.get(
            "/api/captcha/img",
        )
        self.assertEqual(200, resp.status_code)
        self.assertEqual("image/png", resp.headers["content-type"])
        self.assertGreater(len(resp.headers["x-captcha-token"]), 0)

    def test_batch(self):
        base_count = 2
        for i in range(10):
            resp = self.client.post(
                "/api/nodes",
                json={
                    "md": f"node{i}\ntext",
                    "type": const.NodeTypeEnum.MARKDOWN.value,
                },
                headers=self.default_headers,
            )
            self.check_ok_response(resp, 201)

        resp = self.client.get(
            "/api/nodes",
            params={
                "q": "",
                "sort": "createdAt",
                "ord": "desc",
                "p": 0,
                "limit": 5
            },
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        self.assertEqual(5, len(rj["data"]["nodes"]))
        self.assertEqual(10 + base_count, rj["data"]["total"])

        resp = self.client.put(
            "/api/trash/batch",
            json={
                "nids": [n["id"] for n in rj["data"]["nodes"][:3]],
            },
            headers=self.default_headers,
        )
        self.check_ok_response(resp, 200)

        resp = self.client.get(
            "/api/trash",
            params={"p": 0, "limit": 10},
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        self.assertEqual(3, len(rj["data"]["nodes"]))
        self.assertEqual(3, rj["data"]["total"])

        resp = self.client.put(
            "/api/trash/batch/restore",
            json={
                "nids": [n["id"] for n in rj["data"]["nodes"][:2]],
            },
            headers=self.default_headers,
        )
        self.check_ok_response(resp, 200)

        resp = self.client.get(
            "/api/trash",
            params={"p": 0, "limit": 10},
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        self.assertEqual(1, len(rj["data"]["nodes"]))
        self.assertEqual(1, rj["data"]["total"])

        resp = self.client.put(
            "/api/trash/batch/delete",
            json={
                "nids": [n["id"] for n in rj["data"]["nodes"]],
            },
            headers=self.default_headers,
        )
        self.check_ok_response(resp, 200)
        resp = self.client.get(
            "/api/trash",
            params={"p": 0, "limit": 10},
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        self.assertEqual(0, len(rj["data"]["nodes"]))
        self.assertEqual(0, rj["data"]["total"])

    def test_update_obsidian(self):
        image = Image.new('RGB', (4, 4))
        img_byte_arr = io.BytesIO()
        image.save(img_byte_arr, format='PNG')
        orig_data: Dict[str, bytes] = {
            "test1.md": "111\n[[qa]] ![img/p.png](img/p.png)".encode("utf-8"),
            "test2.md": "111\n\nasdq[[test]] ![[Pasted image 20230810112909.png]]".encode("utf-8"),
            "test3.md": "111\n\n![[Pasted image 20230810112931.png]]".encode("utf-8"),
            "20230810112909.png": img_byte_arr.getvalue(),
            os.path.join("img", "p.png"): img_byte_arr.getvalue(),
            os.path.join("img", "20230810112931.png"): img_byte_arr.getvalue(),
        }
        # write dummy zip file
        zip_bytes = ZipFile("test.zip", "w")
        for k, v in orig_data.items():
            zip_bytes.writestr(k, v)
        zip_bytes.close()
        f = open("test.zip", "rb")

        resp = self.client.post(
            "/api/files/obsidian",
            files=[
                ("files", f),
            ],
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 202)

        for _ in range(10):
            time.sleep(0.1)
            resp = self.client.get(
                "/api/files/upload-process",
                headers=self.default_headers,
            )
            rj = self.check_ok_response(resp, 200)
            self.assertEqual("obsidian", rj["type"], msg=rj)
            self.assertEqual("done", rj["msg"], msg=rj)
            if rj["process"] == 100:
                break
        self.assertFalse(rj["running"], msg=rj)

        resp = self.client.get(
            "/api/nodes",
            params={
                "q": "",
                "sort": "createdAt",
                "ord": "desc",
                "p": 0,
                "limit": 5,
            },
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        self.assertEqual(5, len(rj["data"]["nodes"]))
        self.assertEqual(5, rj["data"]["total"])

        f.close()
        os.remove("test.zip")

    def test_upload_text(self):
        os.makedirs("temp", exist_ok=True)
        f1 = open("temp/test1.txt", "wb+")
        f1.write("dasd".encode("utf-8"))
        f2 = open("temp/test2.txt", "wb+")
        f2.write("asdq".encode("utf-8"))

        resp = self.client.post(
            "/api/files/text",
            files=[
                ("files", f1),
                ("files", f2),
            ],
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 202)

        for _ in range(10):
            time.sleep(0.1)
            resp = self.client.get(
                "/api/files/upload-process",
                headers=self.default_headers,
            )
            self.assertEqual(200, resp.status_code)
            rj = resp.json()
            self.assertEqual("md", rj["type"], msg=rj)
            self.assertEqual("done", rj["msg"], msg=rj)
            if rj["process"] == 100:
                break
        self.assertFalse(rj["running"], msg=rj)

        resp = self.client.get(
            "/api/nodes",
            params={
                "q": "",
                "sort": "createdAt",
                "ord": "desc",
                "p": 0,
                "limit": 5,
            },
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        self.assertEqual(4, len(rj["data"]["nodes"]))

        f1.close()
        f2.close()
        shutil.rmtree("temp", ignore_errors=True)

    def test_upload_image(self):
        os.makedirs("temp", exist_ok=True)
        image = Image.new('RGB', (4, 4))
        image.save("temp/test.png", format='PNG')
        f1 = open("temp/test.png", "rb")
        resp = self.client.post(
            "/api/files/vditor",
            files={"file[]": f1},
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        self.assertEqual({
            'errFiles': [],
            'succMap': {
                'test.png': "/files/3acca26d4f9d111694d7dbda2d1e6a40.png"
            }}, rj["data"])
        f1.close()
        shutil.rmtree("temp", ignore_errors=True)

    def test_upload_file(self):
        os.makedirs("temp", exist_ok=True)
        f1 = open("temp/test.txt", "wb+")
        f1.write("dasd".encode("utf-8"))
        resp = self.client.post(
            "/api/files/vditor",
            files={"file[]": f1},
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        self.assertEqual({
            'errFiles': [],
            'succMap': {
                'test.txt': "/files/196b0f14eba66e10fba74dbf9e99c22f.txt"
            }}, rj["data"])
        f1.close()
        shutil.rmtree("temp", ignore_errors=True)

    def test_upload_invalid_file(self):
        os.makedirs("temp", exist_ok=True)
        f1 = open("temp/test.qw", "wb+")
        f1.write("dasd".encode("utf-8"))
        resp = self.client.post(
            "/api/files/vditor",
            files={"file[]": f1},
            headers=self.default_headers,
        )
        self.assertEqual(200, resp.status_code, msg=resp.json())
        rj = resp.json()
        self.assertEqual(const.CodeEnum.INVALID_FILE_TYPE.value, rj["code"])
        self.assertEqual({'errFiles': ['test.qw'], 'succMap': {}}, rj["data"])
        f1.close()
        shutil.rmtree("temp", ignore_errors=True)

    def test_fetch_image(self):
        img = "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAUAAAAFCAYAAACNbyblAAAAHElEQVQI12P4//8/" \
              "w38GIAXDIBKE0DHxgljNBAAO9TXL0Y4OHwAAAABJRU5ErkJggg=="
        resp = self.client.post(
            "/api/files/vditor/images",
            json={"url": img},
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        self.assertEqual(img, rj["data"]["url"])

        img = "fffew"
        resp = self.client.post(
            "/api/files/vditor/images",
            json={"url": img},
            headers=self.default_headers,
        )
        self.error_check(resp, 400, const.CodeEnum.FILE_OPEN_ERROR)

    @patch(
        "retk.utils.httpx.AsyncClient.get",
        return_value=Response(200, content="<title>百度一下</title>".encode("utf-8"))
    )
    def test_put_quick_node(self, mocker):
        resp = self.client.post(
            "/api/nodes/quick",
            json={
                "md": "node1\ntext",
                "type": const.NodeTypeEnum.MARKDOWN.value,
            },
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 201)
        node = rj["node"]
        self.assertEqual("node1", node["title"])
        self.assertEqual("node1\ntext", node["md"])
        self.assertEqual(const.NodeTypeEnum.MARKDOWN.value, node["type"])

        resp = self.client.post(
            "/api/nodes/quick",
            json={
                "md": "https://baidu.com",
                "type": const.NodeTypeEnum.MARKDOWN.value,
            },
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 201)
        node = rj["node"]
        self.assertIn("https://baidu.com", node["md"])
        self.assertIn("百度", node["md"])

    @patch("retk.core.node.backup.__remove_md_all_versions_from_cos")
    @patch("retk.core.node.backup.__remove_md_from_cos")
    @patch("retk.core.node.backup.__get_md_from_cos")
    @patch("retk.core.node.backup.__save_md_to_cos")
    def test_md_history(
            self,
            mock_save_md_to_cos,
            mock_get_md_from_cos,
            mock_remove_md_from_cos,
            mock_remove_md_all_versions_from_cos,
    ):
        mock_save_md_to_cos.return_value = const.CodeEnum.OK
        mock_get_md_from_cos.return_value = ("title2\ntext", const.CodeEnum.OK)
        mock_remove_md_from_cos.return_value = const.CodeEnum.OK
        mock_remove_md_all_versions_from_cos.return_value = const.CodeEnum.OK

        bi = config.get_settings().MD_BACKUP_INTERVAL
        config.get_settings().MD_BACKUP_INTERVAL = 0.0001
        resp = self.client.post(
            "/api/nodes",
            json={
                "md": "title\ntext",
                "type": const.NodeTypeEnum.MARKDOWN.value,
            },
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 201)
        n1 = rj["node"]

        time.sleep(0.001)

        resp = self.client.put(
            f"/api/nodes/{n1['id']}/md",
            json={
                "md": "title1\ntext",
            },
            headers=self.default_headers,
        )
        self.check_ok_response(resp, 200)

        time.sleep(0.001)

        resp = self.client.put(
            f"/api/nodes/{n1['id']}/md",
            json={
                "md": "title2\ntext",
            },
            headers=self.default_headers,
        )
        self.check_ok_response(resp, 200)

        resp = self.client.get(
            f"/api/nodes/{n1['id']}/history",
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        hist = rj["versions"]
        self.assertEqual(2, len(hist))

        resp = self.client.get(
            f"/api/nodes/{n1['id']}/history/{hist[1]}/md",
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        self.assertEqual("title1\ntext", rj["md"])

        config.get_settings().MD_BACKUP_INTERVAL = bi

    @patch("retk.plugins.base.Plugin.handle_api_call")
    def test_plugin(self, mock_handle_api_call):
        def check_one_plugin(ps):
            self.assertGreater(len(ps), 1)
            p = ps[0]
            self.assertIn("id", p)
            self.assertIn("name", p)
            self.assertIn("version", p)
            self.assertIn("description", p)
            self.assertIn("author", p)
            self.assertIn("iconSrc", p)

        config.get_settings().PLUGINS = True
        register_official_plugins()
        resp = self.client.get(
            "/api/plugins",
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        check_one_plugin(rj["plugins"])

        resp = self.client.get(
            "/api/plugins/editor-side",
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        check_one_plugin(rj["plugins"])
        pid = rj["plugins"][0]["id"]

        resp = self.client.get(
            f"/api/plugins/{pid}",
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        self.assertNotEqual("", rj["html"])

        resp = self.client.get(
            "/api/nodes",
            params={
                "q": "",
                "sort": "createdAt",
                "ord": "desc",
                "p": 0,
                "limit": 5,
            },
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        nid = rj["data"]["nodes"][0]["id"]

        resp = self.client.get(
            f"/api/plugins/{pid}/editor-side/{nid}",
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        self.assertNotEqual("", rj["html"])

        resp = self.client.post(
            "/api/plugins/call",
            json={
                "requestId": "xxx",
                "pluginId": "xasdqw",
                "method": "test",
                "data": "test",
            },
            headers=self.default_headers,
        )
        rj = resp.json()
        self.assertEqual(False, rj["success"])

        mock_handle_api_call.return_value = PluginAPICallReturn(
            success=True,
            message="",
            data="test"
        )
        resp = self.client.post(
            "/api/plugins/call",
            json={
                "pluginId": pid,
                "method": "method",
                "data": "test",
            },
            headers=self.default_headers,
        )
        rj = resp.json()
        self.assertEqual("method", rj["method"])
        self.assertEqual("test", rj["data"])

        unregister_official_plugins()
        config.get_settings().PLUGINS = False

    async def test_manager(self):
        manager_token = self.client.cookies.get(const.settings.COOKIE_ACCESS_TOKEN)
        resp = self.client.put(
            "/api/managers/users/disable",
            json={
                "uid": "xxx",
            },
            headers=self.default_headers,
        )
        self.error_check(resp, 403, const.CodeEnum.NOT_PERMITTED)

        admin_uid = await self.set_default_manager()

        resp = self.client.put(
            "/api/managers/users/disable",
            json={
                "uid": "xxx",
            },
            headers=self.default_headers,
        )
        self.error_check(resp, 404, const.CodeEnum.USER_NOT_EXIST)

        email = "a@b.cd"
        resp = await self.create_new_temp_user(email)

        u_token = resp.cookies.get(const.settings.COOKIE_ACCESS_TOKEN)
        uid = (await client.coll.users.find_one({"email": email}))["id"]

        self.set_access_token(manager_token)
        resp = self.client.put(
            "/api/managers/users",
            json={
                "uid": uid,
            },
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        self.assertEqual(email, rj["user"]["email"])
        self.assertEqual(uid, rj["user"]["id"])
        self.assertIn("source", rj["user"])

        resp = self.client.put(
            "/api/managers/users",
            json={
                "email": email,
            },
            headers=self.default_headers,
        )
        self.check_ok_response(resp, 200)

        resp = self.client.put(
            "/api/managers/users",
            json={},
            headers=self.default_headers,
        )
        self.error_check(resp, 400, const.CodeEnum.INVALID_PARAMS)

        resp = self.client.put(
            "/api/managers/users/disable",
            json={
                "uid": uid,
            },
            headers=self.default_headers,
        )
        self.check_ok_response(resp, 200)

        resp = self.client.put(
            "/api/account/login",
            json={
                "email": email,
                "password": "abc111",
            },
            headers=self.default_headers,
        )
        self.error_check(resp, 403, const.CodeEnum.USER_DISABLED)

        self.set_access_token(u_token)
        resp = self.client.get(
            "/api/users",
            headers=self.default_headers,
        )
        self.error_check(resp, 404, const.CodeEnum.USER_NOT_EXIST)

        self.set_access_token(manager_token)
        resp = self.client.put(
            "/api/managers/users/enable",
            json={
                "email": email,
            },
            headers=self.default_headers,
        )
        self.check_ok_response(resp, 200)

        self.set_access_token(u_token)
        resp = self.client.get(
            "/api/users",
            headers=self.default_headers,
        )
        self.check_ok_response(resp, 200)

        self.set_access_token(manager_token)
        resp = self.client.put(
            "/api/managers/users/delete",
            json={
                "uid": uid,
            },
            headers=self.default_headers,
        )
        self.check_ok_response(resp, 200)

        self.set_access_token(u_token)
        resp = self.client.get(
            "/api/users",
            headers=self.default_headers,
        )
        self.error_check(resp, 404, const.CodeEnum.USER_NOT_EXIST)

        await self.clear_default_manager(admin_uid)

    async def test_statistic_user_behavior(self):
        # login
        self.client.put(
            "/api/account/login",
            json={
                "email": const.DEFAULT_USER["email"],
                "password": "",
            }
        )
        uid = (await client.coll.users.find_one({"email": const.DEFAULT_USER["email"]})).get("id")
        docs = await client.coll.user_behavior.find(
            {"uid": uid}
        ).to_list(None)
        self.assertEqual(const.UserBehaviorTypeEnum.LOGIN.value, docs[-1]["type"])

        # create node
        resp = self.client.post(
            "/api/nodes",
            json={
                "md": "node1\ntext",
                "type": const.NodeTypeEnum.MARKDOWN.value,
            },
            headers=self.default_headers,
        )
        docs = await client.coll.user_behavior.find(
            {"uid": uid}
        ).to_list(None)
        self.assertEqual(const.UserBehaviorTypeEnum.NODE_CREATE.value, docs[-1]["type"])
        self.assertEqual(resp.json()["node"]["id"], docs[-1]["remark"])

        # quick node
        resp = self.client.post(
            "/api/nodes/quick",
            json={
                "md": "node1\ntext",
                "type": const.NodeTypeEnum.MARKDOWN.value,
            },
            headers=self.default_headers,
        )
        docs = await client.coll.user_behavior.find(
            {"uid": uid}
        ).to_list(None)
        self.assertEqual(const.UserBehaviorTypeEnum.NODE_QUICK_CREATE.value, docs[-1]["type"])
        self.assertEqual(resp.json()["node"]["id"], docs[-1]["remark"])

        # trash node
        self.client.put(
            f"/api/trash/{resp.json()['node']['id']}",
            headers=self.default_headers,
        )
        docs = await client.coll.user_behavior.find(
            {"uid": uid}
        ).to_list(None)
        self.assertEqual(const.UserBehaviorTypeEnum.NODE_TRASHED_OPS.value, docs[-1]["type"])

        # restore node
        self.client.put(
            f"/api/trash/{resp.json()['node']['id']}/restore",
            headers=self.default_headers,
        )
        docs = await client.coll.user_behavior.find(
            {"uid": uid}
        ).to_list(None)
        self.assertEqual(const.UserBehaviorTypeEnum.NODE_RESTORED_OPS.value, docs[-1]["type"])

        # delete node
        self.client.put(
            f"/api/trash/{resp.json()['node']['id']}",
            headers=self.default_headers,
        )
        self.client.delete(
            f"/api/trash/{resp.json()['node']['id']}",
            headers=self.default_headers,
        )
        docs = await client.coll.user_behavior.find(
            {"uid": uid}
        ).to_list(None)
        self.assertEqual(const.UserBehaviorTypeEnum.NODE_DELETED_OPS.value, docs[-1]["type"])

        # search
        self.client.get(
            "/api/nodes",
            params={
                "q": "1",
                "sort": "createdAt",
                "ord": "desc",
                "p": 0,
                "limit": 5
            },
            headers=self.default_headers,
        )
        docs = await client.coll.user_behavior.find(
            {"uid": uid}
        ).to_list(None)
        self.assertEqual(const.UserBehaviorTypeEnum.SEARCH_GLOBAL.value, docs[-1]["type"])
        self.assertEqual("1", docs[-1]["remark"])

        # search at
        self.client.get(
            f"/api/nodes/{resp.json()['node']['id']}/at",
            params={
                "q": "node",
                "p": 0,
                "limit": 10,
            },
            headers=self.default_headers,
        )
        docs = await client.coll.user_behavior.find(
            {"uid": uid}
        ).to_list(None)
        self.assertEqual(const.UserBehaviorTypeEnum.SEARCH_AT.value, docs[-1]["type"])
        self.assertEqual("node", docs[-1]["remark"])

        # logout
        resp = self.client.post(
            "/api/statistic/user-behavior",
            json={
                "type": const.UserBehaviorTypeEnum.LOGOUT.value,
                "remark": "logout",
            },
            headers=self.default_headers,
        )
        self.assertEqual(201, resp.status_code)
        docs = await client.coll.user_behavior.find(
            {"uid": uid}
        ).to_list(None)
        self.assertEqual(const.UserBehaviorTypeEnum.LOGOUT.value, docs[-1]["type"])
        self.assertEqual("logout", docs[-1]["remark"])

    async def test_system_notice(self):
        manager_token = self.client.cookies.get(const.settings.COOKIE_ACCESS_TOKEN)

        email = "a@b.cd"
        resp = await self.create_new_temp_user(email)
        u_token = resp.cookies.get(const.settings.COOKIE_ACCESS_TOKEN)
        self.assertEqual(201, resp.status_code)

        admin_uid = await self.set_default_manager()
        pa = datetime.datetime.now(tz=utc)
        resp = self.client.post(
            "/api/managers/notices/system",
            json={
                "title": "title",
                "content": "content",
                "recipientType": const.notice.RecipientTypeEnum.ALL.value,
                "batchTypeIds": [],
                "publishAt": pa.strftime("%Y-%m-%dT%H:%M:%SZ"),
            },
            headers=self.default_headers,
        )
        self.error_check(resp, 403, const.CodeEnum.NOT_PERMITTED)

        self.set_access_token(manager_token)
        resp = self.client.post(
            "/api/managers/notices/system",
            json={
                "title": "title",
                "content": "content",
                "recipientType": const.notice.RecipientTypeEnum.ALL.value,
                "batchTypeIds": [],
                "publishAt": pa.strftime("%Y-%m-%dT%H:%M:%SZ"),
            },
            headers=self.default_headers,
        )
        self.check_ok_response(resp, 201)

        resp = self.client.get(
            "/api/managers/notices/system",
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        self.assertEqual(1, rj["total"])
        self.assertEqual(1, len(rj["notices"]))
        self.assertEqual("title", rj["notices"][0]["title"])
        self.assertEqual("<p>content</p>", rj["notices"][0]["html"])
        self.assertEqual("content", rj["notices"][0]["snippet"])
        self.assertEqual(pa.strftime("%Y-%m-%dT%H:%M:%SZ"), rj["notices"][0]["publishAt"])
        self.assertFalse(rj["notices"][0]["scheduled"])

        scheduler.run_once_now(
            job_id="deliver_unscheduled_system_notices1",
            func=scheduler.tasks.notice.deliver_unscheduled_system_notices,
        )

        docs = await client.coll.notice_manager_delivery.find(
            {"senderId": admin_uid}
        ).to_list(None)
        self.assertEqual(1, len(docs))
        self.assertEqual(const.notice.RecipientTypeEnum.ALL.value, docs[0]["recipientType"])
        self.assertEqual("title", docs[0]["title"])
        self.assertEqual("<p>content</p>", docs[0]["html"])
        self.assertEqual(admin_uid, docs[0]["senderId"])
        self.assertEqual([], docs[0]["batchTypeIds"])
        self.assertFalse(docs[0]["scheduled"])
        self.assertEqual(pa.second, docs[0]["publishAt"].second)

        j = scheduler.get_job("deliver_unscheduled_system_notices1")
        for _ in range(10):
            if j.finished_at is not None:
                break
            time.sleep(0.1)
        self.assertIsNotNone(j.finished_at)

        resp = self.client.get(
            "/api/managers/notices/system",
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        self.assertTrue(rj["notices"][0]["scheduled"])

        self.set_access_token(u_token)
        resp = self.client.get(
            "/api/users/notices",
            headers=self.default_headers,
        )
        self.check_ok_response(resp, 200)
        rj = resp.json()
        self.assertIn("system", rj)
        system_notices = rj["system"]["notices"]
        self.assertEqual(1, len(system_notices))
        sn = system_notices[0]
        self.assertEqual("title", sn["title"])
        self.assertEqual("content", sn["snippet"])

        resp = self.client.get(
            f"/api/notices/system/{sn['id']}",
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        self.assertEqual("title", rj["notice"]["title"])
        self.assertEqual("<p>content</p>", rj["notice"]["html"])

        await self.clear_default_manager(admin_uid)

    async def test_user_notice(self):
        manager_token = self.client.cookies.get(const.settings.COOKIE_ACCESS_TOKEN)

        email = "a@b.cd"
        resp = await self.create_new_temp_user(email)
        u_token = resp.cookies.get(const.settings.COOKIE_ACCESS_TOKEN)
        self.assertEqual(201, resp.status_code)

        admin_uid = await self.set_default_manager()
        self.set_access_token(manager_token)
        for i in range(3):
            pa = datetime.datetime.now(tz=utc) - datetime.timedelta(seconds=i + 1)
            resp = self.client.post(
                "/api/managers/notices/system",
                json={
                    "title": f"title{i}",
                    "content": f"content{i}",
                    "recipientType": const.notice.RecipientTypeEnum.ALL.value,
                    "batchTypeIds": [],
                    "publishAt": pa.strftime("%Y-%m-%dT%H:%M:%SZ"),
                },
                headers=self.default_headers,
            )
            self.check_ok_response(resp, 201)

        resp = self.client.get(
            "/api/managers/notices/system",
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        self.assertEqual(3, rj["total"])
        self.assertEqual(3, len(rj["notices"]))

        scheduler.run_once_now(
            job_id="deliver_unscheduled_system_notices2",
            func=scheduler.tasks.notice.deliver_unscheduled_system_notices,
        )
        j = scheduler.get_job("deliver_unscheduled_system_notices2")
        for _ in range(10):
            if j.finished_at is not None:
                break
            time.sleep(0.1)
        self.assertIsNotNone(j.finished_at)
        self.assertEqual("send success 6/6 users", j.finished_return)

        self.set_access_token(u_token)
        resp = self.client.get(
            "/api/users/notices",
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        self.assertEqual(3, rj["system"]["total"])
        self.assertEqual(3, len(rj["system"]["notices"]))
        for i, sn in enumerate(rj["system"]["notices"]):
            self.assertEqual(f"title{2 - i}", sn["title"])
            self.assertEqual(f"content{2 - i}", sn["snippet"])
            self.assertFalse(sn["read"])
            self.assertIsNone(sn["readTime"])

        read_id = rj["system"]['notices'][0]['id']
        resp = self.client.put(
            f"/api/users/notices/system/read/{read_id}",
            headers=self.default_headers,
        )
        self.check_ok_response(resp, 200)

        resp = self.client.get(
            "/api/users/notices",
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        self.assertEqual(3, rj["system"]["total"])
        self.assertEqual(3, len(rj["system"]["notices"]))
        for i, sn in enumerate(rj["system"]["notices"]):
            self.assertEqual(f"title{2 - i}", sn["title"])
            self.assertEqual(f"content{2 - i}", sn["snippet"])
            if read_id == sn["id"]:
                self.assertTrue(sn["read"])
                self.assertIsNotNone(sn["readTime"])
            else:
                self.assertFalse(sn["read"])
                self.assertIsNone(sn["readTime"])

        resp = self.client.put(
            "/api/users/notices/system/read-all",
            headers=self.default_headers,
        )
        self.check_ok_response(resp, 200)

        resp = self.client.get(
            "/api/users/notices",
            headers=self.default_headers,
        )
        rj = self.check_ok_response(resp, 200)
        self.assertEqual(3, rj["system"]["total"])
        self.assertEqual(3, len(rj["system"]["notices"]))
        for i, sn in enumerate(rj["system"]["notices"]):
            self.assertEqual(f"title{2 - i}", sn["title"])
            self.assertEqual(f"content{2 - i}", sn["snippet"])
            self.assertTrue(sn["read"])
            self.assertIsNotNone(sn["readTime"])

        await self.clear_default_manager(admin_uid)



from fastapi import APIRouter, Query

from retk.controllers import schemas, manager
from retk.routes import utils

router = APIRouter(
    prefix="/api/managers",
    tags=["admin"],
    responses={404: {"description": "Not found"}},
)

ADMIN_AUTH = utils.ANNOTATED_AUTHED_ADMIN


@router.get(
    "/",
    status_code=200,
    response_model=schemas.manager.GetManagerDataResponse,
)
@utils.measure_time_spend
async def get_manager_data(
        au: ADMIN_AUTH,
) -> schemas.manager.GetManagerDataResponse:
    return await manager.get_manager_data(au=au)


@router.put(
    "/users",
    status_code=200,
    response_model=schemas.manager.GetUserResponse,
)
@utils.measure_time_spend
async def get_user_info(
        au: ADMIN_AUTH,
        req: schemas.manager.GetUserRequest,
) -> schemas.manager.GetUserResponse:
    return await manager.get_user_info(au=au, req=req)


@router.put(
    "/users/disable",
    status_code=200,
    response_model=schemas.RequestIdResponse,
)
@utils.measure_time_spend
async def disable_account(
        au: ADMIN_AUTH,
        req: schemas.manager.GetUserRequest,
) -> schemas.RequestIdResponse:
    return await manager.disable_account(au=au, req=req)


@router.put(
    "/users/enable",
    status_code=200,
    response_model=schemas.RequestIdResponse,
)
@utils.measure_time_spend
async def enable_account(
        au: ADMIN_AUTH,
        req: schemas.manager.GetUserRequest,
) -> schemas.RequestIdResponse:
    return await manager.enable_account(au=au, req=req)


@router.put(
    "/users/delete",
    status_code=200,
    response_model=schemas.RequestIdResponse,
)
@utils.measure_time_spend
async def delete_account(
        au: ADMIN_AUTH,
        req: schemas.manager.GetUserRequest,
) -> schemas.RequestIdResponse:
    return await manager.delete_account(au=au, req=req)


@router.post(
    "/notices/system",
    status_code=201,
    response_model=schemas.RequestIdResponse,
    summary="Post in manager delivery",
    description="Post a notice in manager delivery",
)
@utils.measure_time_spend
async def post_in_manager_delivery(
        au: utils.ANNOTATED_AUTHED_ADMIN,
        req: schemas.manager.ManagerNoticeDeliveryRequest,
) -> schemas.RequestIdResponse:
    return await manager.post_in_manager_delivery(au=au, req=req)


@router.get(
    "/notices/system",
    status_code=200,
    response_model=schemas.manager.GetSystemNoticesResponse,
    summary="Get unscheduled system notices",
    description="Get unscheduled system notices",
)
@utils.measure_time_spend
async def get_unscheduled_system_notices(
        au: ADMIN_AUTH,
        p: int = Query(0, ge=0),
        limit: int = Query(10, ge=1, le=100),
) -> schemas.manager.GetSystemNoticesResponse:
    return await manager.get_system_notices(au=au, page=p, limit=limit)



from typing import Tuple, Dict

from retk import const
from retk.models.client import client


async def get_marco_data() -> Tuple[Dict, const.CodeEnum]:
    user_count = await client.coll.users.count_documents({})
    node_count = await client.coll.nodes.count_documents({})
    return {
        "user_count": user_count,
        "node_count": node_count,
    }, const.CodeEnum.OK



from fastapi import APIRouter

from retk.controllers import plugin as plugin_ops
from retk.controllers import schemas
from retk.routes import utils

router = APIRouter(
    prefix="/api/plugins",
    tags=["plugins"],
    responses={404: {"description": "Not found"}},
)


@router.get(
    path="/",
    status_code=200,
    response_model=schemas.plugin.PluginsResponse,
)
@utils.measure_time_spend
async def get_plugin_home(
        au: utils.ANNOTATED_AUTHED_USER,
) -> schemas.plugin.PluginsResponse:
    return await plugin_ops.get_all_plugins(au=au)


# must before /{pid}, otherwise /editor-side will be treated as /{pid} as a string
@router.get(
    path="/editor-side",
    status_code=200,
    response_model=schemas.plugin.PluginsResponse,
)
@utils.measure_time_spend
async def get_plugin_with_editor_side(
        au: utils.ANNOTATED_AUTHED_USER,
) -> schemas.plugin.PluginsResponse:
    return await plugin_ops.get_plugins_with_render_editor_side(au=au)


@router.get(
    path="/{pid}",
    status_code=200,
    response_model=schemas.plugin.RenderPluginResponse,
)
@utils.measure_time_spend
async def render_plugin_home(
        au: utils.ANNOTATED_AUTHED_USER,
        pid: str = utils.ANNOTATED_PID,
) -> schemas.plugin.RenderPluginResponse:
    return await plugin_ops.render_plugin_home(au=au, pid=pid)


@router.get(
    path="/{pid}/editor-side/{nid}",
    status_code=200,
    response_model=schemas.plugin.RenderPluginResponse,
)
@utils.measure_time_spend
async def render_editor_side(
        au: utils.ANNOTATED_AUTHED_USER,
        pid: str = utils.ANNOTATED_PID,
        nid: str = utils.ANNOTATED_NID,
) -> schemas.plugin.RenderPluginResponse:
    return await plugin_ops.render_editor_side(au=au, pid=pid, nid=nid)


@router.post(
    path="/call",
    status_code=200,
    response_model=schemas.plugin.PluginCallResponse
)
@utils.measure_time_spend
async def plugin_call(
        req: schemas.plugin.PluginCallRequest,
) -> schemas.plugin.PluginCallResponse:
    return await plugin_ops.plugin_call(req=req)
